import click
from pathlib import Path

import secrets
import shutil
import gradio as gr
from gradio_molecule3d import Molecule3D
from Bio.PDB import PDBIO

from alphafold3_pytorch import Alphafold3, Alphafold3Input

# constants
model = None
cache_path = None
pdb_writer = PDBIO()


# main fold function
def fold(entities, request: gr.Request):
    proteins = []
    rnas = []
    dnas = []
    ligands = []
    ions = []
    for entity in entities:
        if entity["mol_type"] == "Protein":
            proteins.extend([entity["sequence"]] * entity["num_copies"])
        elif entity["mol_type"] == "RNA":
            rnas.extend([entity["sequence"]] * entity["num_copies"])
        elif entity["mol_type"] == "DNA":
            dnas.extend([entity["sequence"]] * entity["num_copies"])
        elif entity["mol_type"] == "Ligand":
            ligands.extend([entity["sequence"]] * entity["num_copies"])
        elif entity["mol_type"] == "Ion":
            ions.extend([entity["sequence"]] * entity["num_copies"])

    # Prepare the input for the model
    alphafold3_input = Alphafold3Input(
        proteins=proteins,
        ss_dna=dnas,
        ss_rna=rnas,
        ligands=ligands,
        metal_ions=ions,
    )

    # Run the model inference in a separate thread
    model.eval()
    (structure,) = model.forward_with_alphafold3_inputs(
        alphafold3_inputs=alphafold3_input,
        return_bio_pdb_structures=True,
    )

    global cache_path, pdb_writer
    output_path = cache_path / str(request.session_hash) / f"{secrets.token_urlsafe(8)}.pdb"
    output_path.parent.mkdir(exist_ok=True)

    pdb_writer.set_structure(structure)
    pdb_writer.save(str(output_path))

    return str(output_path)


# gradio
def delete_cache(request: gr.Request):
    if not request.session_hash:
        return

    user_dir: Path = cache_path / request.session_hash
    if user_dir.exists():
        shutil.rmtree(str(user_dir))


with gr.Blocks(delete_cache=(600, 3600)) as gradio_app:
    entities = gr.State([])

    with gr.Row():
        gr.Markdown("### AlphaFold3 PyTorch Web UI")

    with gr.Row():
        gr.Column(scale=8)
        # upload_json_button = gr.Button("Upload JSON", scale=1, min_width=100)
        clear_button = gr.Button("Clear", scale=1, min_width=100)

    with gr.Row():
        with gr.Column(scale=1, min_width=150):
            mtype = gr.Dropdown(
                value="Protein",
                label="Molecule type",
                choices=["Protein", "DNA", "RNA", "Ligand", "Ion"],
                interactive=True,
            )
        with gr.Column(scale=1, min_width=80):
            c = gr.Number(
                value=1,
                label="Copies",
                interactive=True,
            )

        with gr.Column(scale=8, min_width=200):

            @gr.render(inputs=mtype)
            def render_sequence(mol_type):
                if mol_type in ["Protein", "DNA", "RNA"]:
                    seq = gr.Textbox(
                        label="Paste sequence or fasta",
                        placeholder="Input",
                        interactive=True,
                    )
                elif mol_type == "Ligand":
                    seq = gr.Dropdown(
                        label="Select ligand",
                        choices=[
                            "ADP - Adenosine disphosphate",
                            "ATP - Adenosine triphosphate",
                            "AMP - Adenosine monophosphate",
                            "GTP - Guanosine-5'-triphosphate",
                            "GDP - Guanosine-5'-diphosphate",
                            "FAD - Flavin adenine dinucleotide",
                            "NAD - Nicotinamide-adenine-dinucleotide",
                            "NAP - Nicotinamide-adenine-dinucleotide phosphate (NADP)",
                            "NDP - Dihydro-nicotinamide-adenine-dinucleotide-phosphate (NADPH)",
                            "HEM - Heme",
                            "HEC - Heme C",
                            "OLA - Oleic acid",
                            "MYR - Myristic acid",
                            "CIT - Citric acid",
                            "CLA - Chlorophyll A",
                            "CHL - Chlorophyll B",
                            "BCL - Bacteriochlorophyll A",
                            "BCB - Bacteriochlorophyll B",
                        ],
                        interactive=True,
                    )
                elif mol_type == "Ion":
                    seq = gr.Dropdown(
                        label="Select ion",
                        choices=[
                            "Mg²⁺",
                            "Zn²⁺",
                            "Cl⁻",
                            "Ca²⁺",
                            "Na⁺",
                            "Mn²⁺",
                            "K⁺",
                            "Fe³⁺",
                            "Cu²⁺",
                            "Co²⁺",
                        ],
                        interactive=True,
                    )

                add_button.click(add_entity, inputs=[entities, mtype, c, seq], outputs=[entities])
                clear_button.click(lambda: ("Protein", 1, None), None, outputs=[mtype, c, seq])

    add_button = gr.Button("Add entity", scale=1, min_width=100)

    def add_entity(entities, mtype="Protein", c=1, seq=""):
        if seq is None or len(seq) == 0:
            gr.Info("Input required")
            return entities

        seq_norm = seq.strip(" \t\n\r").upper()

        if mtype in ["Protein", "DNA", "RNA"]:
            if mtype == "Protein" and any([x not in "ARDCQEGHILKMNFPSTWYV" for x in seq_norm]):
                gr.Info("Invalid protein sequence. Allowed characters: A, R, D, C, Q, E, G, H, I, L, K, M, N, F, P, S, T, W, Y, V")
                return entities

            if mtype == "DNA" and any([x not in "ACGT" for x in seq_norm]):
                gr.Info("Invalid DNA sequence. Allowed characters: A, C, G, T")
                return entities

            if mtype == "RNA" and any([x not in "ACGU" for x in seq_norm]):
                gr.Info("Invalid RNA sequence. Allowed characters: A, C, G, U")
                return entities

            if len(seq) < 4:
                gr.Info("Minimum 4 characters required")
                return entities

        elif mtype == "Ligand":
            if seq is None or len(seq) == 0:
                gr.Info("Select a ligand")
                return entities
            seq_norm = seq.split(" - ")[0]
        elif mtype == "Ion":
            if seq is None or len(seq) == 0:
                gr.Info("Select an ion")
                return entities
            seq_norm = "".join([x for x in seq if x.isalpha()])

        new_entity = {"mol_type": mtype, "num_copies": c, "sequence": seq_norm}

        return entities + [new_entity]

    @gr.render(inputs=entities)
    def render_entities(entity_list):
        for idx, entity in enumerate(entity_list):
            with gr.Row():
                gr.Text(
                    value=entity["mol_type"],
                    label="Type",
                    scale=1,
                    min_width=90,
                    interactive=False,
                )
                gr.Text(
                    value=entity["num_copies"],
                    label="Copies",
                    scale=1,
                    min_width=80,
                    interactive=False,
                )

                sequence = entity["sequence"]
                if entity["mol_type"] not in ["Ligand", "Ion"]:
                    # Split every 10 characters, and add a \t after each split
                    sequence = "\t".join([sequence[i : i + 10] for i in range(0, len(sequence), 10)])

                gr.Text(
                    value=sequence,
                    label="Sequence",
                    placeholder="Input",
                    scale=7,
                    min_width=200,
                    interactive=False,
                )

                del_button = gr.Button("🗑️", scale=0, min_width=50)

                def delete(entity_id=idx):
                    entity_list.pop(entity_id)
                    return entity_list

                del_button.click(delete, None, outputs=[entities])

    pred_button = gr.Button("Predict", scale=1, min_width=100)
    output_mol = Molecule3D(label="Output structure", config={"backgroundColor": "black"})

    pred_button.click(fold, inputs=entities, outputs=output_mol)
    clear_button.click(lambda: ([], None), None, outputs=[entities, output_mol])

    gradio_app.unload(delete_cache)


# cli
@click.command()
@click.option("-ckpt", "--checkpoint", type=str, help="path to alphafold3 checkpoint", required=True)
@click.option("-cache", "--cache-dir", type=str, help="path to output cache", required=False, default="cache")
@click.option("-prec", "--precision", type=str, help="precision to use", required=False, default="float32")
def app(checkpoint: str, cache_dir: str, precision: str):
    path = Path(checkpoint)
    assert path.exists(), "checkpoint does not exist at path"

    global cache_path
    cache_path = Path(cache_dir)

    if cache_path.exists():
        shutil.rmtree(str(cache_path))

    cache_path.mkdir(exist_ok=True)

    global model
    model = Alphafold3.init_and_load(str(path))
    # To device and quantize?
    # device = "cuda" if torch.cuda.is_available() else "cpu"
    # try:
    #     dtype = getattr(torch, precision)
    # except AttributeError:
    #     print(f"Invalid precision: {precision}. Using float32")
    #     dtype = torch.float32
    # model.to(device, dtype=dtype)

    gradio_app.launch()



import collections
import copy

import numpy as np
import polars as pl
import scipy
from beartype.typing import Dict, List

from alphafold3_pytorch.tensor_typing import typecheck
from alphafold3_pytorch.data.data_pipeline import GAP_ID
from alphafold3_pytorch.utils.utils import exists

# Constants

MSA_FEATURES = (
    "msa",
    "has_deletion",
    "deletion_value",
)

MSA_PAD_VALUES = {
    "msa_all_seq": GAP_ID,
    "msa_species_identifiers_all_seq": "",
    "has_deletion_all_seq": 0.0,
    "deletion_value_all_seq": 0.0,
    "profile_all_seq": 0.0,
    "deletion_mean_all_seq": 0.0,
    "msa": GAP_ID,
    "has_deletion": 0.0,
    "deletion_value": 0.0,
    "profile": 0.0,
    "deletion_mean": 0.0,
}


@typecheck
def _make_msa_df(chain_features: Dict[str, np.ndarray], gap_id: int = GAP_ID) -> pl.DataFrame:
    """
    Make DataFrame with MSA features needed for MSA pairing.
    From:
    https://github.com/aqlaboratory/openfold/blob/6f63267114435f94ac0604b6d89e82ef45d94484/openfold/data/msa_pairing.py#L118

    :param chain_features: The MSA chain feature dictionary.
    :return: The DataFrame with MSA features.
    """
    chain_msa = chain_features["msa_all_seq"]
    query_seq = chain_msa[0]
    per_seq_similarity = np.sum(query_seq[None] == chain_msa, axis=-1) / float(len(query_seq))
    per_seq_gap = np.sum(chain_msa == gap_id, axis=-1) / float(len(query_seq))
    msa_df = pl.DataFrame(
        {
            "msa_species_identifiers": chain_features["msa_species_identifiers_all_seq"],
            "msa_row": np.arange(len(chain_features["msa_species_identifiers_all_seq"])),
            "msa_similarity": per_seq_similarity,
            "gap": per_seq_gap,
        }
    )
    return msa_df


@typecheck
def _create_species_dict(msa_df: pl.DataFrame) -> Dict[str, pl.DataFrame]:
    """
    Create mapping from species to MSA dataframe of that species.
    From:
    https://github.com/aqlaboratory/openfold/blob/6f63267114435f94ac0604b6d89e82ef45d94484/openfold/data/msa_pairing.py#L137

    :param msa_df: The DataFrame with MSA features.
    :return: The mapping from species to MSA dataframe of that species.
    """
    species_lookup = {}
    for species_tuple, species_df in msa_df.group_by("msa_species_identifiers"):
        species_lookup[species_tuple[0]] = species_df
    return species_lookup


@typecheck
def _match_rows_by_sequence_similarity(
    this_species_msa_dfs: List[pl.DataFrame],
) -> List[np.ndarray]:
    """Find MSA sequence pairings across chains based on sequence similarity.

    Each chain's MSA sequences are first sorted by their sequence similarity to
    their respective target sequence. The sequences are then paired, starting
    from the sequences most similar to their target sequence.

    From:
    https://github.com/aqlaboratory/openfold/blob/6f63267114435f94ac0604b6d89e82ef45d94484/openfold/data/msa_pairing.py#L145

    :param this_species_msa_dfs: A list of DataFrames containing MSA features for sequences
        of a specific species.
    :return: A list of NumPy arrays, each containing M indices corresponding to paired MSA rows,
        where M is the number of chains.
    """
    all_paired_msa_rows = []

    num_seqs = [len(species_df) for species_df in this_species_msa_dfs if exists(species_df)]
    take_num_seqs = np.min(num_seqs)

    for species_df in this_species_msa_dfs:
        if exists(species_df):
            species_df_sorted = species_df.sort("msa_similarity", descending=True)
            msa_rows = species_df_sorted.get_column("msa_row").head(take_num_seqs).to_list()
        else:
            msa_rows = [-1] * take_num_seqs  # Take the last 'padding' row.
        all_paired_msa_rows.append(msa_rows)

    all_paired_msa_rows = list(np.array(all_paired_msa_rows).transpose())
    return all_paired_msa_rows


@typecheck
def pair_sequences(
    chains: List[Dict[str, np.ndarray]],
    max_num_species_sequences: int = 100,
) -> Dict[int, np.ndarray]:
    """
    Return indices for paired MSA sequences across chains.
    From:
    https://github.com/aqlaboratory/openfold/blob/6f63267114435f94ac0604b6d89e82ef45d94484/openfold/data/msa_pairing.py#L181

    :param chains: The MSA chain feature dictionaries.
    :param max_num_species_sequences: The maximum number of sequences to accept for any species.
    :return: The indices for paired MSA sequences across chains.
    """

    num_chains = len(chains)

    all_chain_species_dict = []
    common_species = set()
    for chain_features in chains:
        msa_df = _make_msa_df(chain_features)
        species_dict = _create_species_dict(msa_df)
        all_chain_species_dict.append(species_dict)
        common_species.update(set(species_dict))

    common_species = sorted(common_species)
    common_species.remove("-1")  # Remove target sequence species.

    all_paired_msa_rows = [np.zeros(len(chains), int)]
    all_paired_msa_rows_dict = {k: [] for k in range(num_chains)}
    all_paired_msa_rows_dict[num_chains] = [np.zeros(len(chains), int)]

    for species in common_species:
        if not species:
            continue
        this_species_msa_dfs = []
        species_dfs_present = 0
        for species_dict in all_chain_species_dict:
            if species in species_dict:
                this_species_msa_dfs.append(species_dict[species])
                species_dfs_present += 1
            else:
                this_species_msa_dfs.append(None)

        # Skip species that are present in only one chain.
        if species_dfs_present <= 1:
            continue

        # Skip species with too many sequences.
        if np.any(
            np.array(
                [
                    len(species_df)
                    for species_df in this_species_msa_dfs
                    if isinstance(species_df, pl.DataFrame)
                ]
            )
            > max_num_species_sequences
        ):
            continue

        paired_msa_rows = _match_rows_by_sequence_similarity(this_species_msa_dfs)
        all_paired_msa_rows.extend(paired_msa_rows)
        all_paired_msa_rows_dict[species_dfs_present].extend(paired_msa_rows)

    all_paired_msa_rows_dict = {
        num_examples: np.array(paired_msa_rows)
        for num_examples, paired_msa_rows in all_paired_msa_rows_dict.items()
    }

    return all_paired_msa_rows_dict


@typecheck
def reorder_paired_rows(all_paired_msa_rows_dict: Dict[int, np.ndarray]) -> np.ndarray:
    """Create a list of indices of paired MSA rows across chains.

    :param all_paired_msa_rows_dict: A mapping from the number of paired chains to the paired
        indices.
    :return: A NumPy array, with inner arrays containing indices of paired MSA rows across chains.
        The paired-index lists are ordered by: 1) the number of chains in the paired alignment,
        i.e, all-chain pairings will come first, and 2) e-values
    """
    all_paired_msa_rows = []

    for num_pairings in sorted(all_paired_msa_rows_dict, reverse=True):
        paired_rows = all_paired_msa_rows_dict[num_pairings]
        paired_rows_product = abs(np.array([np.prod(rows) for rows in paired_rows]))
        paired_rows_sort_index = np.argsort(paired_rows_product)
        all_paired_msa_rows.extend(paired_rows[paired_rows_sort_index])

    return np.array(all_paired_msa_rows)


@typecheck
def pad_features(feature: np.ndarray, feature_name: str) -> np.ndarray:
    """Add a 'padding' row at the end of the features list.

    The padding row will be selected as a 'paired' row in the case of partial
    alignment, for the chain that does not have paired alignment.

    From:
    https://github.com/aqlaboratory/openfold/blob/6f63267114435f94ac0604b6d89e82ef45d94484/openfold/data/msa_pairing.py#L91

    :param feature: The feature to be padded.
    :param feature_name: The name of the feature to be padded.
    :return: The feature with an additional padding row.
    """
    assert feature.dtype != np.dtype(np.string_)
    if feature_name in (
        "msa_all_seq",
        "has_deletion_all_seq",
        "deletion_value_all_seq",
    ):
        num_res = feature.shape[1]
        padding = MSA_PAD_VALUES[feature_name] * np.ones([1, num_res], feature.dtype)
    elif feature_name == "msa_species_identifiers_all_seq":
        padding = ["-1"]
    else:
        return feature
    feats_padded = np.concatenate([feature, padding], axis=0)
    return feats_padded


@typecheck
def copy_unpaired_features(chains: List[Dict[str, np.ndarray]]) -> List[Dict[str, np.ndarray]]:
    """Copy unpaired features.

    :param chains: The MSA chain feature dictionaries.
    :return: The MSA chain feature dictionaries with unpaired features copied.
    """
    chain_keys = chains[0].keys()

    new_chains = []
    for chain in chains:
        new_chain_features = copy.deepcopy(chain)
        for feature_name in chain_keys:
            unpaired_feature_name = feature_name.removesuffix("_all_seq")
            if unpaired_feature_name in MSA_FEATURES:
                new_chain_features[unpaired_feature_name] = copy.deepcopy(chain[feature_name])

        new_chains.append(new_chain_features)

    return new_chains


@typecheck
def create_paired_features(
    chains: List[Dict[str, np.ndarray]],
) -> List[Dict[str, np.ndarray]]:
    """
    Pair MSA chain features.
    From:
    https://github.com/aqlaboratory/openfold/blob/6f63267114435f94ac0604b6d89e82ef45d94484/openfold/data/msa_pairing.py#L56

    :param chains: The MSA chain feature dictionaries.
    :return: The MSA chain feature dictionaries with paired features.
    """
    chain_keys = chains[0].keys()

    if len(chains) < 2:
        return chains
    else:
        updated_chains = []
        paired_chains_to_paired_row_indices = pair_sequences(chains)
        paired_rows = reorder_paired_rows(paired_chains_to_paired_row_indices)

        for chain_num, chain in enumerate(chains):
            new_chain = {k: v for k, v in chain.items() if "_all_seq" not in k}

            for feature_name in chain_keys:
                if feature_name in (
                    "profile_all_seq",
                    "deletion_mean_all_seq",
                ):
                    new_chain[feature_name] = chain[feature_name]
                elif feature_name.endswith("_all_seq"):
                    feats_padded = pad_features(chain[feature_name], feature_name)
                    new_chain[feature_name] = feats_padded[paired_rows[:, chain_num]]

            updated_chains.append(new_chain)

        return updated_chains


@typecheck
def merge_homomers_dense_msa(chains: List[Dict[str, np.ndarray]]) -> List[Dict[str, np.ndarray]]:
    """Merge all identical chains, making the resulting MSA dense.

    :param chains: An iterable of features for each chain.
    :return: A list of feature dictionaries.  All features with the same entity_id
        will be merged - MSA features will be concatenated along the num_res
        dimension - making them dense.
    """
    entity_chains = collections.defaultdict(list)
    for chain in chains:
        entity_id = chain["entity_id"][0]
        entity_chains[entity_id].append(chain)

    grouped_chains = []
    for entity_id in sorted(entity_chains):
        chains = entity_chains[entity_id]
        grouped_chains.append(chains)

    chains = [
        merge_features_from_multiple_chains(chains, pair_msa_sequences=True)
        for chains in grouped_chains
    ]
    return chains


@typecheck
def block_diag(*arrs: np.ndarray, pad_value: float = 0.0) -> np.ndarray:
    """Construct a block diagonal like scipy.linalg.block_diag but with an optional padding value.

    :param arrs: The NumPy arrays to block diagonalize.
    :param pad_value: The padding value to use.
    :return: The block diagonalized NumPy array.
    """
    ones_arrs = [np.ones_like(x) for x in arrs]
    off_diag_mask = 1.0 - scipy.linalg.block_diag(*ones_arrs)
    diag = scipy.linalg.block_diag(*arrs)
    diag += (off_diag_mask * pad_value).astype(diag.dtype)
    return diag


@typecheck
def merge_features_from_multiple_chains(
    chains: List[Dict[str, np.ndarray]], pair_msa_sequences: bool
) -> Dict[str, np.ndarray]:
    """Merge features from multiple chains.

    :param chains: A list of feature dictionaries that we want to merge.
    :param pair_msa_sequences: Whether to concatenate MSA features along the num_res dimension (if
        True), or to block diagonalize them (if False).
    :return: A feature dictionary for the merged example.
    """
    chain_keys = chains[0].keys()

    chains_merged = {}
    for feature_name in chain_keys:
        feats = [x[feature_name] for x in chains]
        unpaired_feature_name = feature_name.removesuffix("_all_seq")

        if unpaired_feature_name in MSA_FEATURES:
            if pair_msa_sequences or "_all_seq" in feature_name:
                chains_merged[feature_name] = np.concatenate(feats, axis=-1)
            else:
                chains_merged[feature_name] = block_diag(
                    *feats, pad_value=MSA_PAD_VALUES[feature_name]
                )
        elif feature_name in ("profile_all_seq", "deletion_mean_all_seq"):
            chains_merged[feature_name] = np.concatenate(feats, axis=0)
        else:
            chains_merged[feature_name] = feats[0]

    return chains_merged


@typecheck
def concatenate_paired_and_unpaired_features(
    chains: Dict[str, np.ndarray],
    max_msas_per_chain: int | None = None,
) -> Dict[str, np.ndarray]:
    """Merge paired and unpaired features.

    :param chains: The MSA chain feature dictionaries.
    :param max_msas_per_chain: The maximum number of MSAs per chain.
    :return: The MSA chain feature dictionaries with paired and unpaired features merged.
    """
    if not exists(max_msas_per_chain):
        assert "msa" in chains and "msa_all_seq" in chains, (
            "If max_msas_per_chain is not provided, "
            "both 'msa' and 'msa_all_seq' must be present in 'chains'."
        )
        max_msas_per_chain = max(len(chains["msa"]), len(chains["msa_all_seq"])) * 2

    max_paired_msa_per_chain = max_msas_per_chain // 2

    for feature_name in MSA_FEATURES:
        if feature_name in chains:
            feat = chains[feature_name]
            feat_all_seq = chains[feature_name + "_all_seq"]
            chain_merged = np.concatenate([feat_all_seq[:max_paired_msa_per_chain], feat], axis=0)
            chains[feature_name + "_all_seq"] = chain_merged[:max_msas_per_chain]

            del chains[feature_name]

    return chains




def test_readme1():
    import torch
    from alphafold3_pytorch import Alphafold3
    from alphafold3_pytorch.utils.model_utils import exclusive_cumsum

    alphafold3 = Alphafold3(
        dim_atom_inputs = 77,
        dim_template_feats = 108
    )

    # mock inputs

    seq_len = 16

    molecule_atom_indices = torch.randint(0, 2, (2, seq_len)).long()
    molecule_atom_lens = torch.full((2, seq_len), 2).long()

    atom_seq_len = molecule_atom_lens.sum(dim=-1).amax()
    atom_offsets = exclusive_cumsum(molecule_atom_lens)

    atom_inputs = torch.randn(2, atom_seq_len, 77)
    atompair_inputs = torch.randn(2, atom_seq_len, atom_seq_len, 5)

    additional_molecule_feats = torch.randint(0, 2, (2, seq_len, 5))
    additional_token_feats = torch.randn(2, seq_len, 33)
    is_molecule_types = torch.randint(0, 2, (2, seq_len, 5)).bool()
    is_molecule_mod = torch.randint(0, 2, (2, seq_len, 4)).bool()
    molecule_ids = torch.randint(0, 32, (2, seq_len))

    template_feats = torch.randn(2, 2, seq_len, seq_len, 108)
    template_mask = torch.ones((2, 2)).bool()

    msa = torch.randn(2, 7, seq_len, 32)
    msa_mask = torch.ones((2, 7)).bool()

    additional_msa_feats = torch.randn(2, 7, seq_len, 2)

    # required for training, but omitted on inference

    atom_pos = torch.randn(2, atom_seq_len, 3)

    distogram_atom_indices = molecule_atom_lens - 1

    distance_labels = torch.randint(0, 37, (2, seq_len, seq_len))
    resolved_labels = torch.randint(0, 2, (2, atom_seq_len))

    # offset indices correctly

    distogram_atom_indices += atom_offsets
    molecule_atom_indices += atom_offsets

    # train

    loss = alphafold3(
        num_recycling_steps = 2,
        atom_inputs = atom_inputs,
        atompair_inputs = atompair_inputs,
        molecule_ids = molecule_ids,
        molecule_atom_lens = molecule_atom_lens,
        additional_molecule_feats = additional_molecule_feats,
        additional_msa_feats = additional_msa_feats,
        additional_token_feats = additional_token_feats,
        is_molecule_types = is_molecule_types,
        is_molecule_mod = is_molecule_mod,
        msa = msa,
        msa_mask = msa_mask,
        templates = template_feats,
        template_mask = template_mask,
        atom_pos = atom_pos,
        distogram_atom_indices = distogram_atom_indices,
        molecule_atom_indices = molecule_atom_indices,
        distance_labels = distance_labels,
        resolved_labels = resolved_labels
    )

    loss.backward()

    # after much training ...

    sampled_atom_pos = alphafold3(
        num_recycling_steps = 4,
        num_sample_steps = 16,
        atom_inputs = atom_inputs,
        atompair_inputs = atompair_inputs,
        molecule_ids = molecule_ids,
        molecule_atom_lens = molecule_atom_lens,
        additional_molecule_feats = additional_molecule_feats,
        additional_msa_feats = additional_msa_feats,
        additional_token_feats = additional_token_feats,
        is_molecule_types = is_molecule_types,
        is_molecule_mod = is_molecule_mod,
        msa = msa,
        msa_mask = msa_mask,
        templates = template_feats,
        template_mask = template_mask
    )

def test_readme2():
    import torch
    from alphafold3_pytorch import Alphafold3, Alphafold3Input

    contrived_protein = 'AG'

    mock_atompos = [
        torch.randn(5, 3),   # alanine has 5 non-hydrogen atoms
        torch.randn(4, 3)    # glycine has 4 non-hydrogen atoms
    ]

    train_alphafold3_input = Alphafold3Input(
        proteins = [contrived_protein],
        atom_pos = mock_atompos
    )

    eval_alphafold3_input = Alphafold3Input(
        proteins = [contrived_protein]
    )

    # training

    alphafold3 = Alphafold3(
        dim_atom_inputs = 3,
        dim_atompair_inputs = 5,
        atoms_per_window = 27,
        dim_template_feats = 108,
        num_molecule_mods = 0,
        confidence_head_kwargs = dict(
            pairformer_depth = 1
        ),
        template_embedder_kwargs = dict(
            pairformer_stack_depth = 1
        ),
        msa_module_kwargs = dict(
            depth = 1
        ),
        pairformer_stack = dict(
            depth = 2
        ),
        diffusion_module_kwargs = dict(
            atom_encoder_depth = 1,
            token_transformer_depth = 1,
            atom_decoder_depth = 1,
        )
    )

    loss = alphafold3.forward_with_alphafold3_inputs([train_alphafold3_input])
    loss.backward()

    # sampling

    alphafold3.eval()
    sampled_atom_pos = alphafold3.forward_with_alphafold3_inputs(eval_alphafold3_input)

    assert sampled_atom_pos.shape == (1, (5 + 4), 3)



import os
import shutil
from pathlib import Path

import polars as pl
import pytest

from torch.utils.data import Sampler

from alphafold3_pytorch.data.weighted_pdb_sampler import (
    WeightedPDBSampler
)

from alphafold3_pytorch import (
    create_trainer_from_yaml
)

DATA_TEST_PDB_ID = '209d'

TEST_FOLDER = Path("./data/test/data_caches/clusterings/")

INTERFACE_MAPPING_PATH = str(TEST_FOLDER / "interface_cluster_mapping.csv")

CHAIN_MAPPING_PATHS = [
    str(TEST_FOLDER / "ligand_chain_cluster_mapping.csv"),
    str(TEST_FOLDER / "nucleic_acid_chain_cluster_mapping.csv"),
    str(TEST_FOLDER / "peptide_chain_cluster_mapping.csv"),
    str(TEST_FOLDER / "protein_chain_cluster_mapping.csv"),
]

@pytest.fixture
def sampler():
    """Return a `WeightedPDBSampler` object."""
    return WeightedPDBSampler(
        chain_mapping_paths=CHAIN_MAPPING_PATHS,
        interface_mapping_path=INTERFACE_MAPPING_PATH,
        batch_size=4,
    )


def test_sample(sampler: Sampler):
    """Test the sampling method of the `WeightedSamplerPDB` class."""
    assert len(sampler.sample(4)) == 4, "The sampled batch size does not match the expected size."


def test_cluster_based_sample(sampler: Sampler):
    """Test the cluster-based sampling method of the `WeightedSamplerPDB` class."""
    assert (
        len(sampler.cluster_based_sample(4)) == 4
    ), "The cluster-based sampled batch size does not match the expected size."

# testing end to end with weighted pdb sampler

# testing trainer with pdb inputs

@pytest.fixture()
def populate_mock_pdb_and_remove_test_folders():
    proj_root = Path('.')
    working_cif_file = proj_root / 'data' / 'test' / 'mmcifs' / DATA_TEST_PDB_ID[1:3] / f'{DATA_TEST_PDB_ID}-assembly1.cif'

    pytest_root_folder = Path('./test-folder')
    data_folder = pytest_root_folder / 'data'
    train_folder = data_folder / 'train'
    train_folder.mkdir(exist_ok = True, parents = True)

    pdb_ids = []

    for path in [*CHAIN_MAPPING_PATHS, INTERFACE_MAPPING_PATH]:
        dataset = pl.read_csv(path)
        pdb_ids.extend(list(dataset.get_column('pdb_id')))

    for pdb_id in {*pdb_ids}:
        shutil.copy2(str(working_cif_file), str(train_folder / f'{pdb_id}.cif'))

    yield

    shutil.rmtree('./test-folder')

def test_weighted_sampling_from_trainer_config(populate_mock_pdb_and_remove_test_folders):

    trainer = create_trainer_from_yaml('./tests/configs/trainer_with_pdb_dataset_and_weighted_sampling.yaml')
    trainer()



import numpy as np

from beartype.typing import Any, Iterable, List


def exists(val: Any) -> bool:
    """Check if a value exists.

    :param val: The value to check.
    :return: `True` if the value exists, otherwise `False`.
    """
    return val is not None


def not_exists(val: Any) -> bool:
    """Check if a value does not exist.

    :param val: The value to check.
    :return: `True` if the value does not exist, otherwise `False`.
    """
    return val is None


def default(v: Any, d: Any) -> Any:
    """Return default value `d` if `v` does not exist (i.e., is `None`).

    :param v: The value to check.
    :param d: The default value to return if `v` does not exist.
    :return: The value `v` if it exists, otherwise the default value `d`.
    """
    return v if exists(v) else d


def first(arr: Iterable[Any]) -> Any:
    """Return the first element of an iterable object such as a list.

    :param arr: An iterable object.
    :return: The first element of the iterable object.
    """
    return arr[0]


def always(value):
    """Always return a value."""

    def inner(*args, **kwargs):
        """Inner function."""
        return value

    return inner


def identity(x, *args, **kwargs):
    """Return the input value."""
    return x


def np_mode(x: np.ndarray) -> Any:
    """Return the mode of a 1D NumPy array."""
    assert x.ndim == 1, f"Input NumPy array must be 1D, not {x.ndim}D."
    values, counts = np.unique(x, return_counts=True)
    m = counts.argmax()
    return values[m], counts[m]



from functools import partial

import torch
from torch import tensor
from torch.nn import Module

from beartype.typing import Literal

from alphafold3_pytorch.tensor_typing import (
    typecheck,
    Float,
    Int
)

from alphafold3_pytorch.common.biomolecule import (
    get_residue_constants,
)

from alphafold3_pytorch.inputs import (
    IS_PROTEIN,
)

# class

class ESMWrapper(Module):
    def __init__(
        self,
        esm_name,
        repr_layer = 33
    ):
        super().__init__()
        import esm
        self.repr_layer = repr_layer
        self.model, alphabet = esm.pretrained.load_model_and_alphabet_hub(esm_name)
        self.batch_converter = alphabet.get_batch_converter()

        self.embed_dim = self.model.embed_dim
        self.register_buffer('dummy', tensor(0), persistent = False)

    @torch.no_grad()
    @typecheck
    def forward(
        self,
        aa_ids: Int['b n']
    ) -> Float['b n dpe']:

        device, repr_layer = self.dummy.device, self.repr_layer

        aa_constants = get_residue_constants(res_chem_index=IS_PROTEIN)
        sequence_data = [
            (
                f"molecule{i}",
                "".join(
                    [
                        (
                            aa_constants.restypes[id]
                            if 0 <= id < len(aa_constants.restypes)
                            else "X"
                        )
                        for id in ids
                    ]
                ),
            )
            for i, ids in enumerate(aa_ids)
        ]

        _, _, batch_tokens = self.batch_converter(sequence_data)
        batch_tokens = batch_tokens.to(device)

        self.model.eval()
        results = self.model(batch_tokens, repr_layers=[repr_layer])

        token_representations = results["representations"][repr_layer]

        sequence_representations = []
        for i, (_, seq) in enumerate(sequence_data):
            sequence_representations.append(token_representations[i, 1 : len(seq) + 1])
        plm_embeddings = torch.stack(sequence_representations, dim=0)

        return plm_embeddings

# PLM embedding type and registry

PLMEmbedding = Literal["esm2_t33_650M_UR50D"]

PLMRegistry = dict(
    esm2_t33_650M_UR50D = partial(ESMWrapper, 'esm2_t33_650M_UR50D')
)



"""An mmCIF file format writer."""

import numpy as np

from alphafold3_pytorch.common.biomolecule import _from_mmcif_object, to_mmcif
from alphafold3_pytorch.data.data_pipeline import get_assembly
from alphafold3_pytorch.data.mmcif_parsing import MmcifObject, parse_mmcif_object
from alphafold3_pytorch.utils.utils import exists


def write_mmcif_from_filepath_and_id(
    input_filepath: str, output_filepath: str, file_id: str, **kwargs
):
    """Write an input mmCIF file to an output mmCIF filepath using the provided keyword arguments
    (e.g., sampled coordinates)."""
    mmcif_object = parse_mmcif_object(filepath=input_filepath, file_id=file_id)
    return write_mmcif(mmcif_object, output_filepath=output_filepath, **kwargs)


def write_mmcif(
    mmcif_object: MmcifObject,
    output_filepath: str,
    gapless_poly_seq: bool = True,
    insert_orig_atom_names: bool = True,
    insert_alphafold_mmcif_metadata: bool = True,
    sampled_atom_positions: np.ndarray | None = None,
    b_factors: np.ndarray | None = None,
):
    """Write a BioPython `Structure` object to an mmCIF file using an intermediate `Biomolecule`
    object."""
    biomol = (
        _from_mmcif_object(mmcif_object)
        if "assembly" in mmcif_object.file_id
        else get_assembly(_from_mmcif_object(mmcif_object))
    )
    if exists(sampled_atom_positions):
        atom_mask = biomol.atom_mask.astype(bool)
        assert biomol.atom_positions[atom_mask].shape == sampled_atom_positions.shape, (
            f"Expected sampled atom positions to have masked shape {biomol.atom_positions[atom_mask].shape}, "
            f"but got {sampled_atom_positions.shape}."
        )
        biomol.atom_positions[atom_mask] = sampled_atom_positions
        if exists(b_factors):
            assert biomol.b_factors[atom_mask].shape == b_factors.shape, (
                f"Expected B-factors to have shape {biomol.b_factors[atom_mask].shape}, "
                f"but got {b_factors.shape}."
            )
            biomol.b_factors[atom_mask] = b_factors
    unique_res_atom_names = biomol.unique_res_atom_names if insert_orig_atom_names else None
    mmcif_string = to_mmcif(
        biomol,
        mmcif_object.file_id,
        gapless_poly_seq=gapless_poly_seq,
        insert_alphafold_mmcif_metadata=insert_alphafold_mmcif_metadata,
        unique_res_atom_names=unique_res_atom_names,
    )
    with open(output_filepath, "w") as f:
        f.write(mmcif_string)



import os
import pytest
import shutil
import torch

from alphafold3_pytorch import (
    Alphafold3,
    Alphafold3Input,
    AtomInput,
    AtomDataset,
    PDBInput,
    maybe_transform_to_atom_input,
    collate_inputs_to_batched_atom_input,
    alphafold3_inputs_to_batched_atom_input,
    pdb_inputs_to_batched_atom_input,
    atom_input_to_file,
    file_to_atom_input
)

from alphafold3_pytorch.data import mmcif_writing

from alphafold3_pytorch.life import (
    reverse_complement,
    reverse_complement_tensor
)

from alphafold3_pytorch.mocks import MockAtomDataset

DATA_TEST_PDB_ID = '7a4d'

# reverse complements

def test_string_reverse_complement():
    assert reverse_complement('ATCG') == 'CGAT'
    assert reverse_complement('AUCG', 'rna') == 'CGAU'

def test_tensor_reverse_complement():
    seq = torch.randint(0, 5, (100,))
    rc = reverse_complement_tensor(seq)
    assert torch.allclose(reverse_complement_tensor(rc), seq)

# atom input

def test_atom_input_to_file_and_from():
    mock_atom_dataset = MockAtomDataset(64)
    atom_input = mock_atom_dataset[0]

    file = atom_input_to_file(atom_input, './test-atom-input.pt', overwrite = True)
    atom_input_reconstituted = file_to_atom_input(str(file))
    assert torch.allclose(atom_input.atom_inputs, atom_input_reconstituted.atom_inputs)

def test_atom_dataset():
    num_atom_inputs = 10
    test_folder = './test_atom_folder'

    mock_atom_dataset = MockAtomDataset(num_atom_inputs)

    for i in range(num_atom_inputs):
        atom_input = mock_atom_dataset[i]
        atom_input_to_file(atom_input, f'{test_folder}/{i}.pt', overwrite = True)

    atom_dataset_from_disk = AtomDataset(test_folder)
    assert len(atom_dataset_from_disk) == num_atom_inputs

    shutil.rmtree(test_folder, ignore_errors = True)

# alphafold3 input

@pytest.mark.parametrize('directed_bonds', (False, True))
def test_alphafold3_input(
    directed_bonds
):

    alphafold3_input = Alphafold3Input(
        proteins = ['MLEICLKLVGCKSKKGLSSSSSCYLEEALQRPVASDF', 'MGKCRGLRTARKLRSHRRDQKWHDKQYKKAHLGTALKANPFGGASHAKGIVLEKVGVEAKQPNSAIRKCVRVQLIKNGKKITAFVPNDGCLNFIEENDEVLVAGFGRKGHAVGDIPGVRFKVVKVANVSLLALYKGKKERPRS'],
        ds_dna = ['ACGTT'],
        ds_rna = ['GCCAU', 'CCAGU'],
        ss_dna = ['GCCTA'],
        ss_rna = ['CGCAUA'],
        metal_ions = ['Na', 'Na', 'Fe'],
        misc_molecule_ids = ['Phospholipid'],
        ligands = ['CC1=C(C=C(C=C1)NC(=O)C2=CC=C(C=C2)CN3CCN(CC3)C)NC4=NC=CC(=N4)C5=CN=CC=C5'],
        add_atom_ids = True,
        add_atompair_ids = True,
        directed_bonds = directed_bonds
    )

    batched_atom_input = alphafold3_inputs_to_batched_atom_input(alphafold3_input)

    # feed it into alphafold3

    num_atom_bond_types = (6 * (2 if directed_bonds else 1))

    alphafold3 = Alphafold3(
        dim_atom_inputs = 3,
        dim_atompair_inputs = 5,
        num_atom_embeds = 47,
        num_atompair_embeds = num_atom_bond_types + 1, # 0 is for no bond
        atoms_per_window = 27,
        dim_template_feats = 108,
        num_dist_bins = 64,
        num_molecule_mods = 0,
        confidence_head_kwargs = dict(
            pairformer_depth = 1
        ),
        template_embedder_kwargs = dict(
            pairformer_stack_depth = 1
        ),
        msa_module_kwargs = dict(
            depth = 1
        ),
        pairformer_stack = dict(
            depth = 2
        ),
        diffusion_module_kwargs = dict(
            atom_encoder_depth = 1,
            token_transformer_depth = 1,
            atom_decoder_depth = 1,
        )
    )

    alphafold3(**batched_atom_input.model_forward_dict(), num_sample_steps = 1)

def test_return_bio_pdb_structures():

    alphafold3_input = Alphafold3Input(
        proteins = ['MLEICLKLVGCKSKKGLSSSSSCYLEEALQRPVASDF', 'MGKCRGLRTARKLRSHRRDQKWHDKQYKKAHLGTALKANPFGGASHAKGIVLEKVGVEAKQPNSAIRKCVRVQLIKNGKKITAFVPNDGCLNFIEENDEVLVAGFGRKGHAVGDIPGVRFKVVKVANVSLLALYKGKKERPRS'],
    )

    batched_atom_input = alphafold3_inputs_to_batched_atom_input(alphafold3_input)

    # feed it into alphafold3

    alphafold3 = Alphafold3(
        dim_atom_inputs = 3,
        dim_atompair_inputs = 5,
        num_atom_embeds = 0,
        num_atompair_embeds = 0,
        atoms_per_window = 27,
        dim_template_feats = 108,
        num_dist_bins = 64,
        num_molecule_mods = 0,
        confidence_head_kwargs = dict(
            pairformer_depth = 1
        ),
        template_embedder_kwargs = dict(
            pairformer_stack_depth = 1
        ),
        msa_module_kwargs = dict(
            depth = 1
        ),
        pairformer_stack = dict(
            depth = 2
        ),
        diffusion_module_kwargs = dict(
            atom_encoder_depth = 1,
            token_transformer_depth = 1,
            atom_decoder_depth = 1,
        )
    )

    alphafold3(**batched_atom_input.model_forward_dict(), num_sample_steps = 1, return_bio_pdb_structures = True)

def test_atompos_input():

    contrived_protein = 'AG'
    contrived_nucleic_acid = 'C'

    mock_atompos = [
        torch.randn(5, 3),   # alanine has 5 non-hydrogen atoms
        torch.randn(4, 3),   # glycine has 4 non-hydrogen atoms
        torch.randn(21, 3),  # cytosine has 21 atoms
        torch.randn(3, 3)    # ligand has 3 carbons
    ]

    train_alphafold3_input = Alphafold3Input(
        proteins = [contrived_protein],
        ss_rna = [contrived_nucleic_acid],
        missing_atom_indices = [[1, 2], None, [0, 1], None, None],
        ligands = ['CCC'],
        atom_pos = mock_atompos
    )

    eval_alphafold3_input = Alphafold3Input(
        proteins = [contrived_protein],
        ss_rna = [contrived_nucleic_acid],
        ligands = ['CCC'],
    )

    batched_atom_input = alphafold3_inputs_to_batched_atom_input(train_alphafold3_input, atoms_per_window = 27)

    # training

    alphafold3 = Alphafold3(
        dim_atom_inputs = 3,
        dim_atompair_inputs = 5,
        atoms_per_window = 27,
        dim_template_feats = 108,
        num_dist_bins = 64,
        num_molecule_mods = 0,
        confidence_head_kwargs = dict(
            pairformer_depth = 1
        ),
        template_embedder_kwargs = dict(
            pairformer_stack_depth = 1
        ),
        msa_module_kwargs = dict(
            depth = 1
        ),
        pairformer_stack = dict(
            depth = 2
        ),
        diffusion_module_kwargs = dict(
            atom_encoder_depth = 1,
            token_transformer_depth = 1,
            atom_decoder_depth = 1,
        )
    )

    loss = alphafold3(**batched_atom_input.model_forward_dict())
    loss.backward()

    # sampling

    batched_eval_atom_input = alphafold3_inputs_to_batched_atom_input(eval_alphafold3_input, atoms_per_window = 27)

    alphafold3.eval()
    sampled_atom_pos = alphafold3(**batched_eval_atom_input.model_forward_dict(), return_loss=False)

    assert sampled_atom_pos.shape == (1, (5 + 4 + 21 + 3), 3)

def test_pdbinput_input():
    """Test the PDBInput class, particularly its input transformations for mmCIF files."""
    filepath = os.path.join("data", "test", "mmcifs", DATA_TEST_PDB_ID[1:3], f"{DATA_TEST_PDB_ID}-assembly1.cif")
    file_id = os.path.splitext(os.path.basename(filepath))[0]
    assert os.path.exists(filepath), f"File {filepath} does not exist."

    train_pdb_input = PDBInput(
        filepath,
        chains=("A", "B"),
        cropping_config={
            "contiguous_weight": 0.2,
            "spatial_weight": 0.4,
            "spatial_interface_weight": 0.4,
            "n_res": 64,
        },
        training=True,
    )

    eval_pdb_input = PDBInput(filepath)

    batched_atom_input = pdb_inputs_to_batched_atom_input(train_pdb_input, atoms_per_window=27)

    # training

    alphafold3 = Alphafold3(
        dim_atom=2,
        dim_atompair=2,
        dim_input_embedder_token=8,
        dim_single=2,
        dim_pairwise=2,
        dim_token=2,
        dim_atom_inputs=3,
        dim_atompair_inputs=5,
        atoms_per_window=27,
        dim_template_feats=108,
        num_molecule_mods=4,
        num_dist_bins=64,
        num_rollout_steps=2,
        diffusion_num_augmentations=2,
        confidence_head_kwargs=dict(pairformer_depth=1),
        template_embedder_kwargs=dict(pairformer_stack_depth=1),
        msa_module_kwargs=dict(depth=1, dim_msa=2),
        pairformer_stack=dict(
            depth=1,
            pair_bias_attn_dim_head=1,
            pair_bias_attn_heads=2,
        ),
        diffusion_module_kwargs=dict(
            atom_encoder_depth=1,
            token_transformer_depth=1,
            atom_decoder_depth=1,
            atom_decoder_kwargs=dict(attn_pair_bias_kwargs=dict(dim_head=1)),
            atom_encoder_kwargs=dict(attn_pair_bias_kwargs=dict(dim_head=1)),
        ),
    )

    loss = alphafold3(**batched_atom_input.model_forward_dict())
    loss.backward()

    # sampling

    batched_eval_atom_input = pdb_inputs_to_batched_atom_input(eval_pdb_input, atoms_per_window=27)

    alphafold3.eval()

    batch_dict = batched_eval_atom_input.model_forward_dict()
    sampled_atom_pos = alphafold3(
        **batch_dict,
        return_loss=False,
    )

    batched_atom_mask = ~batch_dict["missing_atom_mask"]
    sampled_atom_positions = sampled_atom_pos[batched_atom_mask].cpu().numpy()

    assert sampled_atom_positions.shape == (4155, 3)

    # visualizing

    mmcif_writing.write_mmcif_from_filepath_and_id(
        input_filepath=filepath,
        output_filepath=filepath.replace(".cif", "-sampled.cif"),
        file_id=file_id,
        gapless_poly_seq=True,
        insert_orig_atom_names=True,
        insert_alphafold_mmcif_metadata=True,
        sampled_atom_positions=sampled_atom_positions,
    )
    assert os.path.exists(filepath.replace(".cif", "-sampled.cif"))



import os
os.environ['TYPECHECK'] = 'True'
os.environ['DEBUG'] = 'True'
from shutil import rmtree
from pathlib import Path

import torch

from alphafold3_pytorch.cli import cli

from alphafold3_pytorch.alphafold3 import (
    Alphafold3
)

def test_cli():
    alphafold3 = Alphafold3(
        dim_atom_inputs = 3,
        dim_template_feats = 44,
        num_molecule_mods = 0
    )

    checkpoint_path = './test-folder/test-cli-alphafold3.pt'
    alphafold3.save(checkpoint_path, overwrite = True)

    cli([
        '--checkpoint', checkpoint_path,
        '-prot', 'AG',
        '-prot', 'TC',
        '--output',
        './test-folder/output.mmcif'
    ], standalone_mode = False)

    assert Path('./test-folder/output.mmcif').exists()

    rmtree('./test-folder')



import pytest
from pathlib import Path

from alphafold3_pytorch import collate_inputs_to_batched_atom_input
from alphafold3_pytorch.alphafold3 import Alphafold3
from alphafold3_pytorch.inputs import (
    PDBDataset,
    molecule_to_atom_input,
    pdb_input_to_molecule_input,
)
from alphafold3_pytorch.data.weighted_pdb_sampler import WeightedPDBSampler


def test_data_input():
    """Test a PDBDataset constructed using a WeightedPDBSampler."""
    data_test = Path("data", "test")
    data_test_mmcif_dir = data_test / "mmcifs"
    data_test_clusterings_dir = data_test / "data_caches" / "clusterings"

    if not data_test_mmcif_dir.exists():
        pytest.skip(f"The directory `{data_test_mmcif_dir}` is not populated yet.")

    interface_mapping_path = str(data_test_clusterings_dir / "interface_cluster_mapping.csv")
    chain_mapping_paths = [
        str(data_test_clusterings_dir / "ligand_chain_cluster_mapping.csv"),
        str(data_test_clusterings_dir / "nucleic_acid_chain_cluster_mapping.csv"),
        str(data_test_clusterings_dir / "peptide_chain_cluster_mapping.csv"),
        str(data_test_clusterings_dir / "protein_chain_cluster_mapping.csv"),
    ]

    sampler = WeightedPDBSampler(
        chain_mapping_paths=chain_mapping_paths,
        interface_mapping_path=interface_mapping_path,
        batch_size=64,
    )

    sampler_pdb_ids = set(sampler.mappings.get_column("pdb_id").to_list())
    test_ids = set(
        filepath.stem
        for filepath in data_test_mmcif_dir.glob("**/*.cif")
        if filepath.stem in sampler_pdb_ids
    )

    dataset = PDBDataset(
        folder=data_test_mmcif_dir,
        sampler=sampler,
        sample_type="default",
        sample_only_pdb_ids=test_ids,
        crop_size=128,
    )

    mol_input = pdb_input_to_molecule_input(pdb_input=dataset[0])
    atom_input = molecule_to_atom_input(mol_input)
    batched_atom_input = collate_inputs_to_batched_atom_input([atom_input], atoms_per_window=27)

    alphafold3 = Alphafold3(
        dim_atom_inputs=3,
        dim_atompair_inputs=5,
        atoms_per_window=27,
        dim_template_feats=108,
        num_dist_bins=64,
        confidence_head_kwargs=dict(pairformer_depth=1),
        template_embedder_kwargs=dict(pairformer_stack_depth=1),
        msa_module_kwargs=dict(depth=1),
        pairformer_stack=dict(depth=2),
        diffusion_module_kwargs=dict(
            atom_encoder_depth=1,
            token_transformer_depth=1,
            atom_decoder_depth=1,
        ),
    )

    loss = alphafold3(**batched_atom_input.model_forward_dict())
    loss.backward()



from __future__ import annotations

import click
from pathlib import Path

import torch

from alphafold3_pytorch import (
    Alphafold3,
    Alphafold3Input
)

from Bio.PDB.mmcifio import MMCIFIO

# simple cli using click

@click.command()
@click.option('-ckpt', '--checkpoint', type = str, help = 'path to alphafold3 checkpoint')
@click.option('-prot', '--protein', type = str, multiple = True, help = 'protein sequences')
@click.option('-rna', '--rna', type = str, multiple = True, help = 'single stranded rna sequences')
@click.option('-dna', '--dna', type = str, multiple = True, help = 'single stranded dna sequences')
@click.option('-o', '--output', type = str, help = 'output path', default = 'output.cif')
def cli(
    checkpoint: str,
    protein: list[str],
    rna: list[str],
    dna: list[str],
    output: str
):

    checkpoint_path = Path(checkpoint)
    assert checkpoint_path.exists(), f'AlphaFold 3 checkpoint must exist at {str(checkpoint_path)}'

    alphafold3_input = Alphafold3Input(
        proteins = protein,
        ss_rna = rna,
        ss_dna = dna,
    )

    alphafold3 = Alphafold3.init_and_load(checkpoint_path)

    alphafold3.eval()
    structure, = alphafold3.forward_with_alphafold3_inputs(alphafold3_input, return_bio_pdb_structures = True)

    output_path = Path(output)
    output_path.parents[0].mkdir(exist_ok = True, parents = True)

    pdb_writer = MMCIFIO()
    pdb_writer.set_structure(structure)
    pdb_writer.save(str(output_path))

    print(f'mmCIF file saved to {str(output_path)}')



"""mmCIF metadata."""

from beartype.typing import Mapping, Sequence

import numpy as np

from importlib.metadata import version

alphafold_version = version("alphafold3_pytorch")

_DISCLAIMER = """THE INFORMATION IS NOT INTENDED FOR, HAS NOT BEEN VALIDATED FOR, AND IS NOT
APPROVED FOR CLINICAL USE. IT SHOULD NOT BE USED FOR CLINICAL PURPOSE OR RELIED
ON FOR MEDICAL OR OTHER PROFESSIONAL ADVICE. IT IS THEORETICAL MODELLING ONLY
AND CAUTION SHOULD BE EXERCISED IN ITS USE. IT IS PROVIDED "AS-IS" WITHOUT ANY
WARRANTY OF ANY KIND, WHETHER EXPRESSED OR IMPLIED. NO WARRANTY IS GIVEN THAT
USE OF THE INFORMATION SHALL NOT INFRINGE THE RIGHTS OF ANY THIRD PARTY."""

# Authors of the Nature methods paper we reference in the mmCIF.
_MMCIF_PAPER_AUTHORS = (
    "Google DeepMind",
    "Isomorphic Labs",
)

# Authors of the mmCIF - we set them to be equal to the authors of the paper.
_MMCIF_AUTHORS = _MMCIF_PAPER_AUTHORS


def add_metadata_to_mmcif(
    old_cif: Mapping[str, Sequence[str]], insert_alphafold_mmcif_metadata: bool = True
) -> Mapping[str, Sequence[str]]:
    """Adds AlphaFold metadata in the given mmCIF."""
    cif = {}

    if insert_alphafold_mmcif_metadata:
        # ModelCIF conformation dictionary.
        cif["_audit_conform.dict_name"] = ["mmcif_ma.dic"]
        cif["_audit_conform.dict_version"] = ["1.4.5"]
        cif["_audit_conform.dict_location"] = [
            "https://raw.githubusercontent.com/ihmwg/ModelCIF/master/dist/" "mmcif_ma.dic"
        ]

        # License and disclaimer.
        cif["_pdbx_data_usage.id"] = ["1", "2"]
        cif["_pdbx_data_usage.type"] = ["license", "disclaimer"]
        cif["_pdbx_data_usage.details"] = [
            "NON-COMMERCIAL USE ONLY, BY USING THIS FILE YOU AGREE TO THE TERMS OF USE FOUND AT alphafoldserver.com/output-terms.",
            _DISCLAIMER,
        ]
        cif["_pdbx_data_usage.url"] = [
            "?",
            "?",
        ]

        # Structure author details.
        cif["_audit_author.name"] = []
        cif["_audit_author.pdbx_ordinal"] = []
        for author_index, author_name in enumerate(_MMCIF_AUTHORS, start=1):
            cif["_audit_author.name"].append(author_name)
            cif["_audit_author.pdbx_ordinal"].append(str(author_index))

        # Paper author details.
        cif["_citation_author.citation_id"] = []
        cif["_citation_author.name"] = []
        cif["_citation_author.ordinal"] = []
        for author_index, author_name in enumerate(_MMCIF_PAPER_AUTHORS, start=1):
            cif["_citation_author.citation_id"].append("primary")
            cif["_citation_author.name"].append(author_name)
            cif["_citation_author.ordinal"].append(str(author_index))

        # Paper citation details.
        cif["_citation.id"] = ["primary"]
        cif["_citation.title"] = [
            "Accurate structure prediction of biomolecular interactions with AlphaFold 3"
        ]
        cif["_citation.journal_full"] = ["Nature"]
        cif["_citation.journal_volume"] = ["?"]
        cif["_citation.page_first"] = ["?"]
        cif["_citation.page_last"] = ["?"]
        cif["_citation.year"] = ["2024"]
        cif["_citation.journal_id_ASTM"] = ["NATUAS"]
        cif["_citation.country"] = ["UK"]
        cif["_citation.journal_id_ISSN"] = ["0028-0836"]
        cif["_citation.journal_id_CSD"] = ["0006"]
        cif["_citation.book_publisher"] = ["?"]
        cif["_citation.pdbx_database_id_PubMed"] = ["?"]
        cif["_citation.pdbx_database_id_DOI"] = ["?"]

        # Type of data in the dataset including data used in the model generation.
        cif["_ma_data.id"] = ["1"]
        cif["_ma_data.name"] = ["Model"]
        cif["_ma_data.content_type"] = ["model coordinates"]

    # Description of number of instances for each entity.
    cif["_ma_target_entity_instance.asym_id"] = old_cif["_struct_asym.id"]
    cif["_ma_target_entity_instance.entity_id"] = old_cif["_struct_asym.entity_id"]
    cif["_ma_target_entity_instance.details"] = ["."] * len(
        cif["_ma_target_entity_instance.entity_id"]
    )

    # Details about the target entities.
    cif["_ma_target_entity.entity_id"] = cif["_ma_target_entity_instance.entity_id"]
    cif["_ma_target_entity.data_id"] = ["1"] * len(cif["_ma_target_entity.entity_id"])
    cif["_ma_target_entity.origin"] = ["."] * len(cif["_ma_target_entity.entity_id"])

    if insert_alphafold_mmcif_metadata:
        # Details of the models being deposited.
        cif["_ma_model_list.ordinal_id"] = ["1"]
        cif["_ma_model_list.model_id"] = ["1"]
        cif["_ma_model_list.model_group_id"] = ["1"]
        cif["_ma_model_list.model_name"] = ["Top ranked model"]

        cif["_ma_model_list.model_group_name"] = [f"AlphaFold v{alphafold_version}"]
        cif["_ma_model_list.data_id"] = ["1"]
        cif["_ma_model_list.model_type"] = ["Ab initio model"]

        # Software used.
        cif["_software.pdbx_ordinal"] = ["1"]
        cif["_software.name"] = ["AlphaFold"]
        cif["_software.version"] = [f"v{alphafold_version}"]
        cif["_software.type"] = ["package"]
        cif["_software.description"] = ["Structure prediction"]
        cif["_software.classification"] = ["other"]
        cif["_software.date"] = ["?"]

        # Collection of software into groups.
        cif["_ma_software_group.ordinal_id"] = ["1"]
        cif["_ma_software_group.group_id"] = ["1"]
        cif["_ma_software_group.software_id"] = ["1"]

        # Method description to conform with ModelCIF.
        cif["_ma_protocol_step.ordinal_id"] = ["1", "2", "3"]
        cif["_ma_protocol_step.protocol_id"] = ["1", "1", "1"]
        cif["_ma_protocol_step.step_id"] = ["1", "2", "3"]
        cif["_ma_protocol_step.method_type"] = [
            "coevolution MSA",
            "template search",
            "modeling",
        ]

        # Details of the metrics use to assess model confidence.
        cif["_ma_qa_metric.id"] = ["1", "2"]
        cif["_ma_qa_metric.name"] = ["pLDDT", "pLDDT"]
        # Accepted values are distance, energy, normalised score, other, zscore.
        cif["_ma_qa_metric.type"] = ["pLDDT", "pLDDT"]
        cif["_ma_qa_metric.mode"] = ["global", "local"]
        cif["_ma_qa_metric.software_group_id"] = ["1", "1"]

        # Global model confidence metric value.
        cif["_ma_qa_metric_global.ordinal_id"] = ["1"]
        cif["_ma_qa_metric_global.model_id"] = ["1"]
        cif["_ma_qa_metric_global.metric_id"] = ["1"]
        global_plddt = np.mean([float(v) for v in old_cif["_atom_site.B_iso_or_equiv"]])
        cif["_ma_qa_metric_global.metric_value"] = [f"{global_plddt:.2f}"]

    cif["_atom_type.symbol"] = sorted(set(old_cif["_atom_site.type_symbol"]))

    return cif



import os
os.environ['TYPECHECK'] = 'True'
os.environ['DEBUG'] = 'True'

import pytest
import random
import itertools
import subprocess
from pathlib import Path

import torch

from einops import repeat
from collections import namedtuple

from alphafold3_pytorch import (
    SmoothLDDTLoss,
    WeightedRigidAlign,
    MultiChainPermutationAlignment,
    ExpressCoordinatesInFrame,
    RigidFrom3Points,
    RigidFromReference3Points,
    ComputeAlignmentError,
    CentreRandomAugmentation,
    PairformerStack,
    MSAModule,
    DiffusionTransformer,
    DiffusionModule,
    ElucidatedAtomDiffusion,
    RelativePositionEncoding,
    TemplateEmbedder,
    Attention,
    InputFeatureEmbedder,
    ConfidenceHead,
    DistogramHead,
    Alphafold3,
    ComputeRankingScore,
    ConfidenceHeadLogits,
    ComputeModelSelectionScore,
    ComputeModelSelectionScore,
    collate_inputs_to_batched_atom_input,
    alphafold3_inputs_to_batched_atom_input,
)

from alphafold3_pytorch.mocks import MockAtomDataset

from alphafold3_pytorch.configs import (
    Alphafold3Config,
    create_alphafold3_from_yaml
)

from alphafold3_pytorch.alphafold3 import (
    mean_pool_with_lens,
    mean_pool_fixed_windows_with_mask,
    batch_repeat_interleave,
    full_pairwise_repr_to_windowed,
    get_cid_molecule_type,
)

from alphafold3_pytorch.inputs import (
    IS_MOLECULE_TYPES,
    IS_PROTEIN,
    atom_ref_pos_to_atompair_inputs,
    molecule_to_atom_input,
    pdb_input_to_molecule_input,
    PDBInput,
    Alphafold3Input,
    PDBDataset,
    default_extract_atom_feats_fn,
    default_extract_atompair_feats_fn,
)

from alphafold3_pytorch.utils.model_utils import (
    exclusive_cumsum,
    get_angle_between_edges,
    get_frames_from_atom_pos,
    get_indices_three_closest_atom_pos,
)

DATA_TEST_PDB_ID = '721p'

def test_atom_ref_pos_to_atompair_inputs():
    atom_ref_pos = torch.randn(16, 3)
    atom_ref_space_uid = torch.ones(16).long()

    atompair_inputs = atom_ref_pos_to_atompair_inputs(atom_ref_pos, atom_ref_space_uid)

    assert atompair_inputs.shape == (16, 16, 5)

def test_mean_pool_with_lens():
    seq = torch.tensor([[[1.], [1.], [1.], [2.], [2.], [2.], [2.], [1.], [1.]]])
    lens = torch.tensor([[3, 4, 2]]).long()
    pooled = mean_pool_with_lens(seq, lens)

    assert torch.allclose(pooled, torch.tensor([[[1.], [2.], [1.]]]))

def test_mean_pool_with_mask():
    seq = torch.tensor([[[1.], [100.], [1.], [2.], [2.], [100.], [1.], [1.], [100.]]])
    mask = torch.tensor([[True, False, True, True, True, False, True, True, False]])

    pooled, _, inverse_function = mean_pool_fixed_windows_with_mask(seq, mask, window_size = 3, return_mask_and_inverse = True)

    assert inverse_function(pooled).shape == seq.shape
    assert torch.allclose(pooled, torch.tensor([[[1.], [2.], [1.]]]))

def test_batch_repeat_interleave():
    seq = torch.tensor([[[1.], [2.], [4.]], [[1.], [2.], [4.]]])
    lens = torch.tensor([[3, 4, 2], [2, 5, 1]]).long()
    repeated = batch_repeat_interleave(seq, lens)
    assert torch.allclose(repeated, torch.tensor([[[1.], [1.], [1.], [2.], [2.], [2.], [2.], [4.], [4.]], [[1.], [1.], [2.], [2.], [2.], [2.], [2.], [4.], [0.]]]))

def test_smooth_lddt_loss():
    pred_coords = torch.randn(2, 100, 3)
    true_coords = torch.randn(2, 100, 3)
    is_dna = torch.randint(0, 2, (2, 100)).bool()
    is_rna = torch.randint(0, 2, (2, 100)).bool()

    loss_fn = SmoothLDDTLoss()
    loss = loss_fn(pred_coords, true_coords, is_dna, is_rna)

    assert loss.numel() == 1

def test_weighted_rigid_align():
    pred_coords = torch.randn(2, 100, 3)
    weights = torch.rand(2, 100)

    align_fn = WeightedRigidAlign()
    aligned_coords = align_fn(pred_coords, pred_coords, weights)

    # `pred_coords` should match itself without any change after alignment

    rmsd = torch.sqrt(((pred_coords - aligned_coords) ** 2).sum(dim=-1).mean(dim=-1))
    assert (rmsd < 1e-5).all()

    random_augment_fn = CentreRandomAugmentation()
    aligned_coords = align_fn(pred_coords, random_augment_fn(pred_coords), weights)

    # `pred_coords` should match a random augmentation of itself after alignment

    rmsd = torch.sqrt(((pred_coords - aligned_coords) ** 2).sum(dim=-1).mean(dim=-1))
    assert (rmsd < 1e-5).all()

def test_weighted_rigid_align_with_mask():
    pred_coords = torch.randn(2, 100, 3)
    true_coords = torch.randn(2, 100, 3)
    weights = torch.rand(2, 100)
    mask = torch.randint(0, 2, (2, 100)).bool()

    align_fn = WeightedRigidAlign()

    # with mask

    aligned_coords = align_fn(pred_coords, true_coords, weights, mask = mask)

    # do it one sample at a time without make

    all_aligned_coords = []

    for one_mask, one_pred_coords, one_true_coords, one_weight in zip(mask, pred_coords, true_coords, weights):
        one_aligned_coords = align_fn(
            one_pred_coords[one_mask][None, ...],
            one_true_coords[one_mask][None, ...],
            one_weight[one_mask][None, ...]
        )

        all_aligned_coords.append(one_aligned_coords.squeeze(0))

    aligned_coords_without_mask = torch.cat(all_aligned_coords, dim = 0)

    # both ways should come out with about the same results

    assert torch.allclose(aligned_coords[mask], aligned_coords_without_mask, atol=1e-5)

def test_multi_chain_permutation_alignment():
    batch_size = 1
    seq_len = 16
    atom_seq_len = 32

    molecule_atom_indices = torch.randint(0, 2, (batch_size, seq_len)).long()
    molecule_atom_lens = torch.full((batch_size, seq_len), 2).long()

    atom_offsets = exclusive_cumsum(molecule_atom_lens)

    pred_coords = torch.randn(batch_size, atom_seq_len, 3)
    true_coords = torch.randn(batch_size, atom_seq_len, 3)
    weights = torch.rand(batch_size, atom_seq_len)
    mask = torch.randint(0, 2, (batch_size, atom_seq_len)).bool()

    token_bonds = torch.randint(0, 2, (batch_size, seq_len, seq_len)).bool()
    additional_molecule_feats = torch.randint(0, 2, (batch_size, seq_len, 5))
    is_molecule_types = torch.randint(0, 2, (batch_size, seq_len, IS_MOLECULE_TYPES)).bool()

    # ensure the asymmetric and entity IDs are consistently ordered and compatible with one another

    additional_molecule_feats[..., : additional_molecule_feats.shape[1] // 2, 2] = 0
    additional_molecule_feats[..., additional_molecule_feats.shape[1] // 2 :, 2] = 1

    additional_molecule_feats[..., : additional_molecule_feats.shape[1] // 2, 3] = 0

    # offset indices correctly

    molecule_atom_indices += atom_offsets

    align_fn = WeightedRigidAlign()
    permute_fn = MultiChainPermutationAlignment(weighted_rigid_align=align_fn)

    aligned_true_coords = align_fn(pred_coords, true_coords, weights, mask=mask)
    permuted_aligned_true_coords = permute_fn(
        pred_coords=pred_coords,
        true_coords=aligned_true_coords,
        molecule_atom_lens=molecule_atom_lens,
        molecule_atom_indices=molecule_atom_indices,
        token_bonds=token_bonds,
        additional_molecule_feats=additional_molecule_feats,
        is_molecule_types=is_molecule_types,
        mask=mask,
    )

    assert permuted_aligned_true_coords.shape == aligned_true_coords.shape

def test_express_coordinates_in_frame():
    batch_size = 2
    num_coords = 100
    coords = torch.randn(batch_size, num_coords, 3)
    frame = torch.randn(batch_size, num_coords, 3, 3)

    express_fn = ExpressCoordinatesInFrame()
    transformed_coords = express_fn(coords, frame)

    assert transformed_coords.shape == (batch_size, num_coords, 3)

    broadcastable_seq_frame = torch.randn(batch_size, 3, 3)
    transformed_coords = express_fn(coords, broadcastable_seq_frame)

    assert transformed_coords.shape == (batch_size, num_coords, 3)

    broadcastable_batch_and_seq_frame = torch.randn(3, 3)
    transformed_coords = express_fn(coords, broadcastable_batch_and_seq_frame)

    assert transformed_coords.shape == (batch_size, num_coords, 3)

def test_rigid_from_three_points():
    rigid_from_3_points = RigidFrom3Points()

    points = torch.randn(7, 11, 23, 3)
    rotation, _ = rigid_from_3_points((points, points, points))
    assert rotation.shape == (7, 11, 23, 3, 3)

def test_rigid_from_reference_three_points():
    rigid_from_reference_3_points = RigidFromReference3Points()

    points = torch.randn(7, 11, 23, 3)
    rotation, _ = rigid_from_reference_3_points((points, points, points))
    assert rotation.shape == (7, 11, 23, 3, 3)

def test_deriving_frames_for_ligands():
    points = torch.tensor([
        [1., 1., 1.],
        [-99, -99, -99],
        [0, 0, 0],
        [100, 100, 100],
        [-1., -1., -1.],
    ])

    frames = get_frames_from_atom_pos(points, filter_colinear_pos = True)

    assert (frames == -1).all()

    frames = get_frames_from_atom_pos(points, filter_colinear_pos = False)

    assert torch.allclose(frames[2], torch.tensor([0, 2, 4]))

    # test with mask

    mask = torch.tensor([True, True, False, False, False])

    frames = get_frames_from_atom_pos(points, mask = mask)

    assert (frames == -1).all()

def test_compute_alignment_error():
    pred_coords = torch.randn(2, 100, 3)
    pred_frames = torch.randn(2, 100, 3, 3)

    # `pred_coords` should match itself in frame basis

    error_fn = ComputeAlignmentError()
    alignment_errors = error_fn(pred_coords, pred_coords, pred_frames, pred_frames)

    assert alignment_errors.shape == (2, 100, 100)
    assert (alignment_errors.mean(-1) < 1e-3).all()

def test_centre_random_augmentation():
    coords = torch.randn(2, 100, 3)

    augmentation_fn = CentreRandomAugmentation()
    augmented_coords = augmentation_fn(coords)

    assert augmented_coords.shape == coords.shape


@pytest.mark.parametrize('checkpoint', (True, False))
@pytest.mark.parametrize('recurrent_depth', (1, 2))
@pytest.mark.parametrize('enable_attn_softclamp', (True, False))
def test_pairformer(
    checkpoint,
    recurrent_depth,
    enable_attn_softclamp
):
    single = torch.randn(2, 16, 384).requires_grad_()
    pairwise = torch.randn(2, 16, 16, 128).requires_grad_()
    mask = torch.randint(0, 2, (2, 16)).bool()

    pairformer = PairformerStack(
        depth = 4,
        num_register_tokens = 4,
        recurrent_depth = recurrent_depth,
        checkpoint = checkpoint,
        pair_bias_attn_kwargs = dict(
            enable_attn_softclamp = enable_attn_softclamp
        )
    )

    single_out, pairwise_out = pairformer(
        single_repr = single,
        pairwise_repr = pairwise,
        mask = mask
    )

    assert single.shape == single_out.shape
    assert pairwise.shape == pairwise_out.shape

    if checkpoint:
        loss = single_out.sum() + pairwise_out.sum()
        loss.backward()

@pytest.mark.parametrize('checkpoint', (False, True))
def test_msa_module(
    checkpoint
):
    single = torch.randn(2, 16, 384).requires_grad_()
    pairwise = torch.randn(2, 16, 16, 128).requires_grad_()
    msa = torch.randn(2, 7, 16, 32)
    mask = torch.randint(0, 2, (2, 16)).bool()
    msa_mask = torch.randint(0, 2, (2, 7)).bool()
    additional_msa_feats = torch.randn(2, 7, 16, 2)

    msa_module = MSAModule(
        checkpoint = checkpoint,
        max_num_msa = 3 # will randomly select 3 out of the MSAs, accounting for mask, using sample without replacement
    )

    pairwise_out = msa_module(
        msa = msa,
        single_repr = single,
        pairwise_repr = pairwise,
        mask = mask,
        msa_mask = msa_mask,
        additional_msa_feats = additional_msa_feats
    )

    assert pairwise.shape == pairwise_out.shape

    if checkpoint:
        loss = pairwise_out.sum()
        loss.backward()

@pytest.mark.parametrize('checkpoint', (False, True))
@pytest.mark.parametrize('use_linear_attn', (False, True))
@pytest.mark.parametrize('use_colt5_attn', (False, True))
def test_diffusion_transformer(
    checkpoint,
    use_linear_attn,
    use_colt5_attn
):

    single = torch.randn(2, 16, 384).requires_grad_()
    pairwise = torch.randn(2, 16, 16, 128).requires_grad_()
    mask = torch.randint(0, 2, (2, 16)).bool()

    diffusion_transformer = DiffusionTransformer(
        depth = 2,
        heads = 16,
        checkpoint = checkpoint,
        use_linear_attn = use_linear_attn,
        use_colt5_attn = use_colt5_attn
    )

    single_out = diffusion_transformer(
        single,
        single_repr = single,
        pairwise_repr = pairwise,
        mask = mask
    )

    assert single.shape == single_out.shape

    if checkpoint:
        loss = single_out.sum()
        loss.backward()

def test_sequence_local_attn():
    atoms = torch.randn(2, 17, 32)
    attn_bias = torch.randn(2, 17, 17)

    attn = Attention(
        dim = 32,
        dim_head = 16,
        heads = 8,
        window_size = 5
    )

    out = attn(atoms, attn_bias = attn_bias)
    assert out.shape == atoms.shape

@pytest.mark.parametrize('karras_formulation', (True, False))
def test_diffusion_module(karras_formulation):
    seq_len = 16
    atom_seq_len = 32

    molecule_atom_indices = torch.randint(0, 2, (2, seq_len)).long()
    molecule_atom_lens = torch.full((2, seq_len), 2).long()

    atom_offsets = exclusive_cumsum(molecule_atom_lens)

    token_bonds = torch.randint(0, 2, (2, seq_len, seq_len)).bool()

    noised_atom_pos = torch.randn(2, atom_seq_len, 3)
    atom_feats = torch.randn(2, atom_seq_len, 128)
    atompair_feats = torch.randn(2, atom_seq_len, atom_seq_len, 16)
    atom_mask = torch.ones((2, atom_seq_len)).bool()

    times = torch.randn(
        2,
    )
    mask = torch.ones(2, seq_len).bool()
    single_trunk_repr = torch.randn(2, seq_len, 128)
    single_inputs_repr = torch.randn(2, seq_len, 256)

    pairwise_trunk = torch.randn(2, seq_len, seq_len, 128)
    pairwise_rel_pos_feats = torch.randn(2, seq_len, seq_len, 12)

    # offset indices correctly

    molecule_atom_indices += atom_offsets

    diffusion_module = DiffusionModule(
        atoms_per_window=27,
        dim_pairwise_trunk=128,
        dim_pairwise_rel_pos_feats=12,
        atom_encoder_depth=1,
        atom_decoder_depth=1,
        token_transformer_depth=1,
        atom_encoder_kwargs=dict(attn_num_memory_kv=2),
        token_transformer_kwargs=dict(num_register_tokens=2),
    )

    atom_pos_update = diffusion_module(
        noised_atom_pos,
        times=times,
        atom_feats=atom_feats,
        atompair_feats=atompair_feats,
        atom_mask=atom_mask,
        mask=mask,
        single_trunk_repr=single_trunk_repr,
        single_inputs_repr=single_inputs_repr,
        pairwise_trunk=pairwise_trunk,
        pairwise_rel_pos_feats=pairwise_rel_pos_feats,
        molecule_atom_lens=molecule_atom_lens,
    )

    assert noised_atom_pos.shape == atom_pos_update.shape

    edm = ElucidatedAtomDiffusion(
        diffusion_module, karras_formulation=karras_formulation, num_sample_steps=2
    )

    edm_return = edm(
        noised_atom_pos,
        atom_feats=atom_feats,
        atompair_feats=atompair_feats,
        atom_mask=atom_mask,
        mask=mask,
        single_trunk_repr=single_trunk_repr,
        single_inputs_repr=single_inputs_repr,
        pairwise_trunk=pairwise_trunk,
        pairwise_rel_pos_feats=pairwise_rel_pos_feats,
        molecule_atom_lens=molecule_atom_lens,
        molecule_atom_indices=molecule_atom_indices,
        token_bonds=token_bonds,
        add_bond_loss=True,
    )

    assert edm_return.loss.numel() == 1

    sampled_atom_pos = edm.sample(
        atom_mask=atom_mask,
        atom_feats=atom_feats,
        atompair_feats=atompair_feats,
        mask=mask,
        single_trunk_repr=single_trunk_repr,
        single_inputs_repr=single_inputs_repr,
        pairwise_trunk=pairwise_trunk,
        pairwise_rel_pos_feats=pairwise_rel_pos_feats,
        molecule_atom_lens=molecule_atom_lens,
    )

    assert sampled_atom_pos.shape == noised_atom_pos.shape
    
def test_relative_position_encoding():
    additional_molecule_feats = torch.randint(0, 2, (8, 100, 5))

    embedder = RelativePositionEncoding()

    rpe_embed = embedder(
        additional_molecule_feats = additional_molecule_feats
    )

@pytest.mark.parametrize('checkpoint', (False, True))
def test_template_embed(
    checkpoint
):
    template_feats = torch.randn(2, 2, 16, 16, 108)
    template_mask = torch.ones((2, 2)).bool()

    pairwise_repr = torch.randn(2, 16, 16, 128).requires_grad_()
    mask = torch.ones((2, 16)).bool()

    embedder = TemplateEmbedder(
        dim_template_feats = 108,
        checkpoint = checkpoint
    )

    template_embed = embedder(
        templates = template_feats,
        template_mask = template_mask,
        pairwise_repr = pairwise_repr,
        mask = mask
    )

    if checkpoint:
        loss = template_embed.sum()
        loss.backward()

def test_confidence_head():
    seq_len = 16
    atom_seq_len = 32

    molecule_atom_indices = torch.randint(0, 2, (2, seq_len)).long()
    molecule_atom_lens = torch.full((2, seq_len), 2).long()

    single_inputs_repr = torch.randn(2, seq_len, 77)
    single_repr = torch.randn(2, seq_len, 384)
    pairwise_repr = torch.randn(2, seq_len, seq_len, 128)
    mask = torch.ones((2, seq_len)).bool()

    atom_feats = torch.randn(2, atom_seq_len, 64)
    pred_atom_pos = torch.randn(2, atom_seq_len, 3)

    confidence_head = ConfidenceHead(
        dim_single_inputs=77,
        dim_atom=64,
        atompair_dist_bins=torch.linspace(3, 20, 37).tolist(),
        dim_single=384,
        dim_pairwise=128,
    )

    logits = confidence_head(
        single_inputs_repr=single_inputs_repr,
        single_repr=single_repr,
        pairwise_repr=pairwise_repr,
        pred_atom_pos=pred_atom_pos,
        atom_feats=atom_feats,
        molecule_atom_indices=molecule_atom_indices,
        molecule_atom_lens=molecule_atom_lens,
        mask=mask,
    )

    assert logits.pae.shape[-1] == seq_len
    assert logits.pde.shape[-1] == seq_len

    assert logits.plddt.shape[-1] == atom_seq_len
    assert logits.resolved.shape[-1] == atom_seq_len

def test_input_embedder():

    molecule_atom_lens = torch.randint(0, 3, (2, 16))
    atom_seq_len = molecule_atom_lens.sum(dim = -1).amax()
    atom_inputs = torch.randn(2, atom_seq_len, 77)
    atompair_inputs = torch.randn(2, atom_seq_len, atom_seq_len, 5)

    atom_mask = torch.ones((2, atom_seq_len)).bool()
    additional_token_feats = torch.randn(2, 16, 33)
    molecule_ids = torch.randint(0, 32, (2, 16))

    embedder = InputFeatureEmbedder(
        dim_atom_inputs = 77,
    )

    embedder(
        atom_inputs = atom_inputs,
        atom_mask = atom_mask,
        atompair_inputs = atompair_inputs,
        molecule_atom_lens = molecule_atom_lens,
        molecule_ids = molecule_ids,
        additional_token_feats = additional_token_feats
    )

def test_distogram_head():
    pairwise_repr = torch.randn(2, 16, 16, 128)

    distogram_head = DistogramHead(dim_pairwise = 128)

    logits = distogram_head(pairwise_repr)

@pytest.mark.parametrize('window_atompair_inputs', (True, False))
@pytest.mark.parametrize('stochastic_frame_average', (True, False))
@pytest.mark.parametrize('missing_atoms', (True, False))
@pytest.mark.parametrize('calculate_pae', (True, False))
@pytest.mark.parametrize('atom_transformer_intramolecular_attn', (True, False))
@pytest.mark.parametrize('num_molecule_mods', (0, 4))
@pytest.mark.parametrize('distogram_atom_resolution', (True, False))
def test_alphafold3(
    window_atompair_inputs: bool,
    stochastic_frame_average: bool,
    missing_atoms: bool,
    calculate_pae: bool,
    atom_transformer_intramolecular_attn: bool,
    num_molecule_mods: int,
    distogram_atom_resolution: bool
):
    seq_len = 16
    atoms_per_window = 27

    molecule_atom_indices = torch.randint(0, 2, (2, seq_len)).long()
    molecule_atom_lens = torch.full((2, seq_len), 3).long()

    atom_seq_len = molecule_atom_lens.sum(dim = -1).amax()
    atom_offset = exclusive_cumsum(molecule_atom_lens)

    token_bonds = torch.randint(0, 2, (2, seq_len, seq_len)).bool()

    atom_inputs = torch.randn(2, atom_seq_len, 77)

    atompair_inputs = torch.randn(2, atom_seq_len, atom_seq_len, 5)

    if window_atompair_inputs:
        atompair_inputs = full_pairwise_repr_to_windowed(atompair_inputs, window_size = atoms_per_window)

    additional_molecule_feats = torch.randint(0, 2, (2, seq_len, 5))
    additional_token_feats = torch.randn(2, 16, 33)
    is_molecule_types = torch.randint(0, 2, (2, seq_len, IS_MOLECULE_TYPES)).bool()
    molecule_ids = torch.randint(0, 32, (2, seq_len))

    is_molecule_mod = None
    if num_molecule_mods > 0:
        is_molecule_mod = torch.zeros(2, seq_len, num_molecule_mods).uniform_(0, 1) < 0.1

    atom_indices_for_frame = None
    if calculate_pae:
        atom_indices_for_frame = repeat(torch.arange(3), 'c -> b n c', b = 2, n = seq_len).clone()
        atom_indices_for_frame += atom_offset[..., None]

    missing_atom_mask = None
    if missing_atoms:
        missing_atom_mask = torch.randint(0, 2, (2, atom_seq_len)).bool()

    atom_parent_ids = None

    if atom_transformer_intramolecular_attn:
        atom_parent_ids = torch.ones(2, atom_seq_len).long()

    template_feats = torch.randn(2, 2, seq_len, seq_len, 108)
    template_mask = torch.ones((2, 2)).bool()

    msa = torch.randn(2, 7, seq_len, 32)
    msa_mask = torch.ones((2, 7)).bool()

    additional_msa_feats = torch.randn(2, 7, seq_len, 2)

    atom_pos = torch.randn(2, atom_seq_len, 3)
    distogram_atom_indices = molecule_atom_lens - 1

    resolved_labels = torch.randint(0, 2, (2, atom_seq_len))

    # ensure the asymmetric and entity IDs are consistently ordered and compatible with one another

    additional_molecule_feats[..., : additional_molecule_feats.shape[1] // 2, 2] = 0
    additional_molecule_feats[..., additional_molecule_feats.shape[1] // 2 :, 2] = 1

    additional_molecule_feats[..., : additional_molecule_feats.shape[1] // 2, 3] = 0

    # offset indices correctly

    distogram_atom_indices += atom_offset
    molecule_atom_indices += atom_offset

    # alphafold3

    alphafold3 = Alphafold3(
        dim_atom_inputs = 77,
        dim_pairwise = 8,
        dim_single = 8,
        dim_token = 8,
        atoms_per_window = atoms_per_window,
        dim_template_feats = 108,
        num_dist_bins = 64,
        num_molecule_mods = num_molecule_mods,
        confidence_head_kwargs = dict(
            pairformer_depth = 1
        ),
        template_embedder_kwargs = dict(
            pairformer_stack_depth = 1
        ),
        msa_module_kwargs = dict(
            depth = 1,
            dim_msa = 8,
        ),
        pairformer_stack=dict(
            depth=1,
            pair_bias_attn_dim_head = 4,
            pair_bias_attn_heads = 2,
        ),
        diffusion_module_kwargs=dict(
            atom_encoder_depth=1,
            token_transformer_depth=1,
            atom_decoder_depth=1,
            atom_decoder_kwargs = dict(
                attn_pair_bias_kwargs = dict(
                    dim_head = 4
                )
            ),
            atom_encoder_kwargs = dict(
                attn_pair_bias_kwargs = dict(
                    dim_head = 4
                )
            )
        ),
        stochastic_frame_average = stochastic_frame_average,
        distogram_atom_resolution = distogram_atom_resolution
    )

    loss, breakdown = alphafold3(
        num_recycling_steps = 2,
        atom_inputs = atom_inputs,
        molecule_ids = molecule_ids,
        molecule_atom_lens = molecule_atom_lens,
        atom_parent_ids = atom_parent_ids,
        atompair_inputs = atompair_inputs,
        missing_atom_mask = missing_atom_mask,
        atom_indices_for_frame = atom_indices_for_frame,
        is_molecule_types = is_molecule_types,
        is_molecule_mod = is_molecule_mod,
        additional_molecule_feats = additional_molecule_feats,
        additional_msa_feats = additional_msa_feats,
        additional_token_feats = additional_token_feats,
        token_bonds = token_bonds,
        msa = msa,
        msa_mask = msa_mask,
        templates = template_feats,
        template_mask = template_mask,
        atom_pos = atom_pos,
        distogram_atom_indices = distogram_atom_indices,
        molecule_atom_indices = molecule_atom_indices,
        resolved_labels = resolved_labels,
        num_rollout_steps = 1,
        diffusion_add_smooth_lddt_loss = True,
        return_loss_breakdown = True
    )

    loss.backward()

    sampled_atom_pos = alphafold3(
        num_sample_steps = 16,
        atom_inputs = atom_inputs,
        molecule_ids = molecule_ids,
        molecule_atom_lens = molecule_atom_lens,
        is_molecule_mod = is_molecule_mod,
        atompair_inputs = atompair_inputs,
        is_molecule_types = is_molecule_types,
        additional_molecule_feats = additional_molecule_feats,
        additional_msa_feats = additional_msa_feats,
        additional_token_feats = additional_token_feats,
        msa = msa,
        templates = template_feats,
        template_mask = template_mask,
    )

    assert sampled_atom_pos.ndim == 3

def test_alphafold3_without_msa_and_templates():
    seq_len = 16
    atom_seq_len = 32

    molecule_atom_indices = torch.randint(0, 2, (2, seq_len)).long()
    molecule_atom_lens = torch.full((2, seq_len), 2).long()

    atom_inputs = torch.randn(2, atom_seq_len, 77)
    atompair_inputs = torch.randn(2, atom_seq_len, atom_seq_len, 5)
    additional_molecule_feats = torch.randint(0, 2, (2, seq_len, 5))
    additional_token_feats = torch.randn(2, seq_len, 33)
    is_molecule_types = torch.randint(0, 2, (2, seq_len, IS_MOLECULE_TYPES)).bool()
    molecule_ids = torch.randint(0, 32, (2, seq_len))

    atom_pos = torch.randn(2, atom_seq_len, 3)
    distogram_atom_indices = molecule_atom_lens - 1

    resolved_labels = torch.randint(0, 2, (2, atom_seq_len))

    alphafold3 = Alphafold3(
        dim_atom_inputs = 77,
        dim_template_feats = 108,
        num_dist_bins = 64,
        num_molecule_mods = 0,
        checkpoint_trunk_pairformer = True,
        checkpoint_diffusion_module = True,
        confidence_head_kwargs = dict(
            pairformer_depth = 1
        ),
        template_embedder_kwargs = dict(
            pairformer_stack_depth = 1
        ),
        msa_module_kwargs = dict(
            depth = 1
        ),
        pairformer_stack = dict(
            depth = 2
        ),
        diffusion_module_kwargs = dict(
            atom_encoder_depth = 2,
            token_transformer_depth = 2,
            atom_decoder_depth = 2,
        ),
    )

    loss, breakdown = alphafold3(
        num_recycling_steps = 2,
        atom_inputs = atom_inputs,
        molecule_ids = molecule_ids,
        molecule_atom_lens = molecule_atom_lens,
        atompair_inputs = atompair_inputs,
        is_molecule_types = is_molecule_types,
        additional_molecule_feats = additional_molecule_feats,
        additional_token_feats = additional_token_feats,
        atom_pos = atom_pos,
        distogram_atom_indices = distogram_atom_indices,
        resolved_labels = resolved_labels,
        return_loss_breakdown = True
    )

    loss.backward()

def test_alphafold3_force_return_loss():
    seq_len = 16
    atom_seq_len = 32

    molecule_atom_lens = torch.full((2, seq_len), 2).long()
    atom_offsets = exclusive_cumsum(molecule_atom_lens)

    molecule_atom_indices = torch.randint(0, 2, (2, seq_len)).long()
    molecule_atom_indices += atom_offsets

    atom_inputs = torch.randn(2, atom_seq_len, 77)
    atompair_inputs = torch.randn(2, atom_seq_len, atom_seq_len, 5)
    additional_molecule_feats = torch.randint(0, 2, (2, seq_len, 5))
    additional_token_feats = torch.randn(2, seq_len, 33)
    is_molecule_types = torch.randint(0, 2, (2, seq_len, IS_MOLECULE_TYPES)).bool()
    molecule_ids = torch.randint(0, 32, (2, seq_len))

    atom_pos = torch.randn(2, atom_seq_len, 3)

    distogram_atom_indices = molecule_atom_lens - 1
    distogram_atom_indices += atom_offsets

    resolved_labels = torch.randint(0, 2, (2, atom_seq_len))

    alphafold3 = Alphafold3(
        dim_atom_inputs = 77,
        dim_template_feats = 108,
        num_dist_bins = 64,
        num_molecule_mods = 0,
        confidence_head_kwargs = dict(
            pairformer_depth = 1
        ),
        template_embedder_kwargs = dict(
            pairformer_stack_depth = 1
        ),
        msa_module_kwargs = dict(
            depth = 1
        ),
        pairformer_stack = dict(
            depth = 2
        ),
        diffusion_module_kwargs = dict(
            atom_encoder_depth = 1,
            token_transformer_depth = 1,
            atom_decoder_depth = 1,
        ),
    )

    sampled_atom_pos = alphafold3(
        num_recycling_steps = 2,
        atom_inputs = atom_inputs,
        molecule_ids = molecule_ids,
        molecule_atom_lens = molecule_atom_lens,
        atompair_inputs = atompair_inputs,
        is_molecule_types = is_molecule_types,
        additional_molecule_feats = additional_molecule_feats,
        additional_token_feats = additional_token_feats,
        atom_pos = atom_pos,
        distogram_atom_indices = distogram_atom_indices,
        molecule_atom_indices = molecule_atom_indices,
        resolved_labels = resolved_labels,
        return_loss_breakdown = True,
        return_loss = False # force sampling even if labels are given
    )

    assert sampled_atom_pos.ndim == 3

    loss, _ = alphafold3(
        num_recycling_steps = 2,
        atom_inputs = atom_inputs,
        molecule_ids = molecule_ids,
        molecule_atom_lens = molecule_atom_lens,
        atompair_inputs = atompair_inputs,
        is_molecule_types = is_molecule_types,
        additional_molecule_feats = additional_molecule_feats,
        additional_token_feats = additional_token_feats,
        molecule_atom_indices = molecule_atom_indices,
        return_loss_breakdown = True,
        return_loss = True # force returning loss even if no labels given
    )

    assert loss == 0.

def test_alphafold3_force_return_loss_with_confidence_logits():
    seq_len = 16
    atom_seq_len = 32

    molecule_atom_lens = torch.full((2, seq_len), 2).long()
    atom_offsets = exclusive_cumsum(molecule_atom_lens)

    molecule_atom_indices = torch.randint(0, 2, (2, seq_len)).long()
    molecule_atom_indices += atom_offsets

    atom_inputs = torch.randn(2, atom_seq_len, 77)
    atompair_inputs = torch.randn(2, atom_seq_len, atom_seq_len, 5)
    additional_molecule_feats = torch.randint(0, 2, (2, seq_len, 5))
    additional_token_feats = torch.randn(2, seq_len, 33)
    is_molecule_types = torch.randint(0, 2, (2, seq_len, IS_MOLECULE_TYPES)).bool()
    molecule_ids = torch.randint(0, 32, (2, seq_len))

    atom_pos = torch.randn(2, atom_seq_len, 3)
    distogram_atom_indices = molecule_atom_lens - 1
    distogram_atom_indices += atom_offsets

    resolved_labels = torch.randint(0, 2, (2, atom_seq_len))

    alphafold3 = Alphafold3(
        dim_atom_inputs = 77,
        dim_template_feats = 108,
        num_dist_bins = 64,
        num_molecule_mods = 0,
        confidence_head_kwargs = dict(
            pairformer_depth = 1
        ),
        template_embedder_kwargs = dict(
            pairformer_stack_depth = 1
        ),
        msa_module_kwargs = dict(
            depth = 1
        ),
        pairformer_stack = dict(
            depth = 2
        ),
        diffusion_module_kwargs = dict(
            atom_encoder_depth = 1,
            token_transformer_depth = 1,
            atom_decoder_depth = 1,
        ),
    )

    sampled_atom_pos, confidence_head_logits = alphafold3(
        num_recycling_steps = 2,
        atom_inputs = atom_inputs,
        molecule_ids = molecule_ids,
        molecule_atom_lens = molecule_atom_lens,
        atompair_inputs = atompair_inputs,
        is_molecule_types = is_molecule_types,
        additional_molecule_feats = additional_molecule_feats,
        additional_token_feats = additional_token_feats,
        atom_pos = atom_pos,
        distogram_atom_indices = distogram_atom_indices,
        molecule_atom_indices = molecule_atom_indices,
        resolved_labels = resolved_labels,
        return_loss_breakdown = True,
        return_loss = False, # force sampling even if labels are given
        return_confidence_head_logits = True
    )

    assert sampled_atom_pos.ndim == 3

    loss, _ = alphafold3(
        num_recycling_steps = 2,
        atom_inputs = atom_inputs,
        molecule_ids = molecule_ids,
        molecule_atom_lens = molecule_atom_lens,
        atompair_inputs = atompair_inputs,
        is_molecule_types = is_molecule_types,
        additional_molecule_feats = additional_molecule_feats,
        additional_token_feats = additional_token_feats,
        return_loss_breakdown = True,
        return_loss = True # force returning loss even if no labels given
    )

    assert loss == 0.

def test_alphafold3_with_atom_and_bond_embeddings():
    alphafold3 = Alphafold3(
        num_atom_embeds = 7,
        num_atompair_embeds = 3,
        num_molecule_mods = 0,
        dim_atom_inputs = 77,
        dim_template_feats = 108
    )

    # mock inputs

    seq_len = 16
    atom_seq_len = 32

    molecule_atom_lens = torch.full((2, seq_len), 2).long()
    atom_offset = exclusive_cumsum(molecule_atom_lens)

    molecule_atom_indices = torch.randint(0, 2, (2, seq_len)).long()
    molecule_atom_indices += atom_offset

    atom_ids = torch.randint(0, 7, (2, atom_seq_len))
    atompair_ids = torch.randint(0, 3, (2, atom_seq_len, atom_seq_len))

    atom_inputs = torch.randn(2, atom_seq_len, 77)
    atompair_inputs = torch.randn(2, atom_seq_len, atom_seq_len, 5)

    additional_molecule_feats = torch.randint(0, 2, (2, seq_len, 5))
    additional_token_feats = torch.randn(2, seq_len, 33)
    is_molecule_types = torch.randint(0, 2, (2, seq_len, IS_MOLECULE_TYPES)).bool()
    molecule_ids = torch.randint(0, 32, (2, seq_len))

    template_feats = torch.randn(2, 2, seq_len, seq_len, 108)
    template_mask = torch.ones((2, 2)).bool()

    msa = torch.randn(2, 7, seq_len, 32)
    msa_mask = torch.ones((2, 7)).bool()

    additional_msa_feats = torch.randn(2, 7, seq_len, 2)

    # required for training, but omitted on inference

    atom_pos = torch.randn(2, atom_seq_len, 3)
    distogram_atom_indices = molecule_atom_lens - 1 # last atom, as an example
    distogram_atom_indices += atom_offset

    resolved_labels = torch.randint(0, 2, (2, atom_seq_len))

    # train

    loss = alphafold3(
        num_recycling_steps = 2,
        atom_ids = atom_ids,
        atompair_ids = atompair_ids,
        atom_inputs = atom_inputs,
        atompair_inputs = atompair_inputs,
        molecule_ids = molecule_ids,
        molecule_atom_lens = molecule_atom_lens,
        is_molecule_types = is_molecule_types,
        additional_molecule_feats = additional_molecule_feats,
        additional_msa_feats = additional_msa_feats,
        additional_token_feats = additional_token_feats,
        msa = msa,
        msa_mask = msa_mask,
        templates = template_feats,
        template_mask = template_mask,
        atom_pos = atom_pos,
        distogram_atom_indices = distogram_atom_indices,
        molecule_atom_indices = molecule_atom_indices,
        resolved_labels = resolved_labels
    )

    assert loss.numel() == 1

def test_alphafold3_with_plm_embeddings():
    alphafold3 = Alphafold3(
        num_atom_embeds=7,
        num_atompair_embeds=3,
        num_molecule_mods=0,
        dim_atom_inputs=77,
        dim_template_feats=108,
        plm_embeddings="esm2_t33_650M_UR50D",
    )

    # mock inputs

    seq_len = 16
    atom_seq_len = 32

    molecule_atom_indices = torch.randint(0, 2, (2, seq_len)).long()
    molecule_atom_lens = torch.full((2, seq_len), 2).long()

    atom_offsets = exclusive_cumsum(molecule_atom_lens)

    atom_ids = torch.randint(0, 7, (2, atom_seq_len))
    atompair_ids = torch.randint(0, 3, (2, atom_seq_len, atom_seq_len))

    atom_inputs = torch.randn(2, atom_seq_len, 77)
    atompair_inputs = torch.randn(2, atom_seq_len, atom_seq_len, 5)

    additional_molecule_feats = torch.randint(0, 2, (2, seq_len, 5))
    additional_token_feats = torch.randn(2, seq_len, 33)
    is_molecule_types = torch.randint(0, 2, (2, seq_len, IS_MOLECULE_TYPES)).bool()
    molecule_ids = torch.randint(0, 32, (2, seq_len))

    template_feats = torch.randn(2, 2, seq_len, seq_len, 108)
    template_mask = torch.ones((2, 2)).bool()

    msa = torch.randn(2, 7, seq_len, 32)
    msa_mask = torch.ones((2, 7)).bool()

    additional_msa_feats = torch.randn(2, 7, seq_len, 2)

    # required for training, but omitted on inference

    atom_pos = torch.randn(2, atom_seq_len, 3)
    distogram_atom_indices = molecule_atom_lens - 1  # last atom, as an example

    resolved_labels = torch.randint(0, 2, (2, atom_seq_len))

    # offset indices correctly

    distogram_atom_indices += atom_offsets
    molecule_atom_indices += atom_offsets

    # alphafold3

    loss = alphafold3(
        num_recycling_steps=2,
        atom_ids=atom_ids,
        atompair_ids=atompair_ids,
        atom_inputs=atom_inputs,
        atompair_inputs=atompair_inputs,
        molecule_ids=molecule_ids,
        molecule_atom_lens=molecule_atom_lens,
        is_molecule_types=is_molecule_types,
        additional_molecule_feats=additional_molecule_feats,
        additional_msa_feats=additional_msa_feats,
        additional_token_feats=additional_token_feats,
        msa=msa,
        msa_mask=msa_mask,
        templates=template_feats,
        template_mask=template_mask,
        atom_pos=atom_pos,
        distogram_atom_indices=distogram_atom_indices,
        molecule_atom_indices=molecule_atom_indices,
        resolved_labels=resolved_labels,
    )

    assert loss.numel() == 1

# test creation from config

def test_alphafold3_config():
    curr_dir = Path(__file__).parents[0]
    af3_yaml = curr_dir / 'configs/alphafold3.yaml'
    trainer_yml = curr_dir / 'configs/trainer.yaml'

    alphafold3 = create_alphafold3_from_yaml(af3_yaml)
    assert isinstance(alphafold3, Alphafold3)

    alphafold3_from_trainer_yml = create_alphafold3_from_yaml(trainer_yml, 'model')
    assert isinstance(alphafold3_from_trainer_yml, Alphafold3)

# test compute ranking score

def test_compute_ranking_score():

    # mock inputs
    
    batch_size = 2
    seq_len = 16
    atom_seq_len = 32

    molecule_atom_lens = torch.full((2, seq_len), 2).long()

    is_molecule_types = torch.randint(0, 2, (batch_size, seq_len, 5)).bool()
    atom_pos = torch.randn(batch_size, atom_seq_len, 3) * 5
    atom_mask = torch.randint(0, 2, (atom_pos.shape[:-1])).type_as(atom_pos).bool()
    has_frame = torch.randint(0, 2, (batch_size, seq_len)).bool()
    is_modified_residue = torch.randint(0, 2, (batch_size, atom_seq_len))

    pae_logits = torch.randn(batch_size, 64, seq_len, seq_len)
    pde_logits = torch.randn(batch_size, 64, seq_len, seq_len)
    plddt_logits = torch.randn(batch_size, 50, atom_seq_len)
    resolved_logits = torch.randint(0, 2, (batch_size, 2, seq_len))
    confidence_head_logits = ConfidenceHeadLogits(pae_logits, pde_logits, plddt_logits, resolved_logits)
    atom_level_pae_logits = torch.randn(batch_size, 64, atom_seq_len, atom_seq_len)

    chain_length = [random.randint(seq_len // 4, seq_len //2) 
                    for _ in range(batch_size)]

    asym_id = torch.tensor([
        [item for val, count in enumerate([chain_len, seq_len - chain_len]) for item in itertools.repeat(val, count)]
        for chain_len in chain_length
    ]).long()


    compute_ranking_score = ComputeRankingScore()

    full_complex_metric = compute_ranking_score.compute_full_complex_metric(
        confidence_head_logits, asym_id, has_frame, molecule_atom_lens,
        atom_pos, atom_mask, is_molecule_types)

    single_chain_metric = compute_ranking_score.compute_single_chain_metric(
        confidence_head_logits, asym_id, has_frame,)

    interface_metric = compute_ranking_score.compute_interface_metric(
        confidence_head_logits, asym_id, has_frame,
        interface_chains=[(0, 1), (1,)])

    modified_residue_score = compute_ranking_score.compute_modified_residue_score(
        confidence_head_logits, atom_mask, is_modified_residue)
    
    residue_level_ptm_score = compute_ranking_score.compute_confidence_score.compute_ptm(
        pae_logits, asym_id, has_frame
    )

    assert full_complex_metric.numel() == batch_size
    assert single_chain_metric.numel() == batch_size
    assert interface_metric.numel() == batch_size
    assert modified_residue_score.numel() == batch_size
    assert residue_level_ptm_score.numel() == batch_size

def test_model_selection_score():

    # mock inputs
    
    batch_size = 2
    seq_len = 16
    atom_seq_len = 32

    molecule_atom_lens = torch.full((2, seq_len), 2).long()

    atom_pos_true = torch.randn(batch_size, atom_seq_len, 3) * 5
    atom_pos_pred = torch.randn(batch_size, atom_seq_len, 3) * 5
    atom_mask = torch.randint(0, 2, (atom_pos_true.shape[:-1])).type_as(atom_pos_true).bool()
    tok_repr_atm_mask = torch.randint(0, 2, (batch_size, seq_len)).bool()

    dist_logits = torch.randn(batch_size, 64, seq_len, seq_len)
    pde_logits = torch.randn(batch_size, 64, seq_len, seq_len)

    chain_length = [random.randint(seq_len // 4, seq_len //2) 
                    for _ in range(batch_size)]

    asym_id = torch.tensor([
        [item for val, count in enumerate([chain_len, seq_len - chain_len]) for item in itertools.repeat(val, count)]
        for chain_len in chain_length
    ]).long()

    is_molecule_types = torch.zeros_like(asym_id)
    is_molecule_types = torch.nn.functional.one_hot(is_molecule_types, 5).bool()
    
    compute_model_selection_score = ComputeModelSelectionScore()

    gpde = compute_model_selection_score.compute_gpde(
        pde_logits,
        dist_logits,
        compute_model_selection_score.dist_breaks,
        tok_repr_atm_mask,
    )

    weighted_lddt = compute_model_selection_score.compute_weighted_lddt(
        atom_pos_pred,
        atom_pos_true,
        atom_mask,
        asym_id,
        is_molecule_types,
        molecule_atom_lens,
        chains_list = [(0, 1), (1,)],
        is_fine_tuning=False
    )

def test_model_selection_score_end_to_end():

    # prepare two atom inputs for evaluating model selection

    mock_atom_dataset = MockAtomDataset(10)

    atom_inputs = [mock_atom_dataset[0], mock_atom_dataset[1]]
    batched_atom_input = collate_inputs_to_batched_atom_input(atom_inputs, atoms_per_window=27)

    # two models to be selected

    alphafold3_kwargs = dict(
        dim_atom_inputs = 77,
        dim_pairwise = 8,
        dim_single = 8,
        dim_token = 8,
        atoms_per_window = 27,
        dim_template_feats = 108,
        num_dist_bins = 64,
        confidence_head_kwargs = dict(
            pairformer_depth = 1
        ),
        template_embedder_kwargs = dict(
            pairformer_stack_depth = 1
        ),
        msa_module_kwargs = dict(
            depth = 1,
            dim_msa = 8,
        ),
        pairformer_stack=dict(
            depth=1,
            pair_bias_attn_dim_head = 4,
            pair_bias_attn_heads = 2,
        ),
        diffusion_module_kwargs=dict(
            atom_encoder_depth=1,
            token_transformer_depth=1,
            atom_decoder_depth=1,
            atom_decoder_kwargs = dict(
                attn_pair_bias_kwargs = dict(
                    dim_head = 4
                )
            ),
            atom_encoder_kwargs = dict(
                attn_pair_bias_kwargs = dict(
                    dim_head = 4
                )
            )
        ),
    )

    alphafold3_one = Alphafold3(**alphafold3_kwargs)
    alphafold3_two = Alphafold3(**alphafold3_kwargs)

    alphafolds = (alphafold3_one, alphafold3_two)

    # evaluate

    compute_model_selection_score = ComputeModelSelectionScore()

    details = compute_model_selection_score(alphafolds, batched_atom_input, return_details = True)

    best_alphafold_by_lddt = alphafolds[details.best_lddt_index]
    assert isinstance(best_alphafold_by_lddt, Alphafold3)

def test_unresolved_protein_rasa():

    # rest of the test

    mmcif_filepath = os.path.join("data", "test", "mmcifs", DATA_TEST_PDB_ID[1:3], f"{DATA_TEST_PDB_ID}-assembly1.cif")
    pdb_input = PDBInput(mmcif_filepath)

    mol_input = pdb_input_to_molecule_input(pdb_input)
    atom_input = molecule_to_atom_input(mol_input)
    batched_atom_input = collate_inputs_to_batched_atom_input([atom_input], atoms_per_window=27)
    batched_atom_input_dict = batched_atom_input.dict()

    res_idx, token_idx, asym_id, entity_id, sym_id = batched_atom_input_dict['additional_molecule_feats'].unbind(dim = -1)

    cid = 1
    res_chem_index = get_cid_molecule_type(
        cid,
        asym_id[0],
        batched_atom_input_dict['is_molecule_types'][0])

    # only support protein for unresolved protein calculate
    assert res_chem_index == IS_PROTEIN

    unresolved_residue_mask = torch.randint(0, 2, asym_id.shape).bool()

    compute_model_selection_score = ComputeModelSelectionScore()

    if not compute_model_selection_score.can_calculate_unresolved_protein_rasa:
        pytest.skip("mkdssp not available for calculating unresolved protein rasa")

    unresolved_rasa = compute_model_selection_score.compute_unresolved_rasa(
        unresolved_cid=[1],
        unresolved_residue_mask=unresolved_residue_mask,
        asym_id = asym_id,
        molecule_ids=batched_atom_input_dict['molecule_ids'],
        molecule_atom_lens=batched_atom_input_dict['molecule_atom_lens'],
        atom_pos=batched_atom_input_dict['atom_pos'],
        atom_mask=~batched_atom_input_dict['missing_atom_mask'])

def test_readme1():
    alphafold3 = Alphafold3(dim_atom_inputs=77, dim_template_feats=108)

    # mock inputs

    seq_len = 16

    molecule_atom_indices = torch.randint(0, 2, (2, seq_len)).long()
    molecule_atom_lens = torch.full((2, seq_len), 2).long()

    atom_seq_len = molecule_atom_lens.sum(dim=-1).amax()
    atom_offsets = exclusive_cumsum(molecule_atom_lens)

    atom_inputs = torch.randn(2, atom_seq_len, 77)
    atompair_inputs = torch.randn(2, atom_seq_len, atom_seq_len, 5)

    additional_molecule_feats = torch.randint(0, 2, (2, seq_len, 5))
    additional_token_feats = torch.randn(2, seq_len, 33)
    is_molecule_types = torch.randint(0, 2, (2, seq_len, 5)).bool()
    is_molecule_mod = torch.randint(0, 2, (2, seq_len, 4)).bool()
    molecule_ids = torch.randint(0, 32, (2, seq_len))

    template_feats = torch.randn(2, 2, seq_len, seq_len, 108)
    template_mask = torch.ones((2, 2)).bool()

    msa = torch.randn(2, 7, seq_len, 32)
    msa_mask = torch.ones((2, 7)).bool()

    additional_msa_feats = torch.randn(2, 7, seq_len, 2)

    # required for training, but omitted on inference

    atom_pos = torch.randn(2, atom_seq_len, 3)

    distogram_atom_indices = molecule_atom_lens - 1

    distance_labels = torch.randint(0, 37, (2, seq_len, seq_len))
    resolved_labels = torch.randint(0, 2, (2, atom_seq_len))

    # offset indices correctly

    distogram_atom_indices += atom_offsets
    molecule_atom_indices += atom_offsets

    # train

    loss = alphafold3(
        num_recycling_steps=2,
        atom_inputs=atom_inputs,
        atompair_inputs=atompair_inputs,
        molecule_ids=molecule_ids,
        molecule_atom_lens=molecule_atom_lens,
        additional_molecule_feats=additional_molecule_feats,
        additional_msa_feats=additional_msa_feats,
        additional_token_feats=additional_token_feats,
        is_molecule_types=is_molecule_types,
        is_molecule_mod=is_molecule_mod,
        msa=msa,
        msa_mask=msa_mask,
        templates=template_feats,
        template_mask=template_mask,
        atom_pos=atom_pos,
        distogram_atom_indices=distogram_atom_indices,
        molecule_atom_indices=molecule_atom_indices,
        distance_labels=distance_labels,
        resolved_labels=resolved_labels,
    )

    loss.backward()

    # after much training ...

    sampled_atom_pos = alphafold3(
        num_recycling_steps=4,
        num_sample_steps=16,
        atom_inputs=atom_inputs,
        atompair_inputs=atompair_inputs,
        molecule_ids=molecule_ids,
        molecule_atom_lens=molecule_atom_lens,
        additional_molecule_feats=additional_molecule_feats,
        additional_msa_feats=additional_msa_feats,
        additional_token_feats=additional_token_feats,
        is_molecule_types=is_molecule_types,
        is_molecule_mod=is_molecule_mod,
        msa=msa,
        msa_mask=msa_mask,
        templates=template_feats,
        template_mask=template_mask,
    )

    sampled_atom_pos.shape  # (2, <atom_seqlen>, 3)
    assert sampled_atom_pos.ndim == 3

def test_readme2():
    contrived_protein = 'AG'

    mock_atompos = [
        torch.randn(5, 3),   # alanine has 5 non-hydrogen atoms
        torch.randn(4, 3)    # glycine has 4 non-hydrogen atoms
    ]

    train_alphafold3_input = Alphafold3Input(
        proteins = [contrived_protein],
        atom_pos = mock_atompos
    )

    eval_alphafold3_input = Alphafold3Input(
        proteins = [contrived_protein]
    )

    batched_atom_input = alphafold3_inputs_to_batched_atom_input(train_alphafold3_input, atoms_per_window = 27)

    # training

    alphafold3 = Alphafold3(
        dim_atom_inputs = 3,
        dim_atompair_inputs = 5,
        atoms_per_window = 27,
        dim_template_feats = 108,
        num_dist_bins = 64,
        num_molecule_mods = 0,
        confidence_head_kwargs = dict(
            pairformer_depth = 1
        ),
        template_embedder_kwargs = dict(
            pairformer_stack_depth = 1
        ),
        msa_module_kwargs = dict(
            depth = 1
        ),
        pairformer_stack = dict(
            depth = 2
        ),
        diffusion_module_kwargs = dict(
            atom_encoder_depth = 1,
            token_transformer_depth = 1,
            atom_decoder_depth = 1,
        )
    )

    loss = alphafold3(**batched_atom_input.model_forward_dict())
    loss.backward()

    # sampling

    batched_eval_atom_input = alphafold3_inputs_to_batched_atom_input(eval_alphafold3_input, atoms_per_window = 27)

    alphafold3.eval()
    sampled_atom_pos = alphafold3(**batched_eval_atom_input.model_forward_dict())

    assert sampled_atom_pos.shape == (1, (5 + 4), 3)



import pytest
from pathlib import Path

from alphafold3_pytorch.inputs import PDBDataset, pdb_inputs_to_batched_atom_input
from alphafold3_pytorch.data.weighted_pdb_sampler import WeightedPDBSampler
from alphafold3_pytorch.utils.utils import exists


def test_template_loading():
    """Test a template-featurized PDBDataset constructed using a WeightedPDBSampler."""
    data_test = Path("data", "test")
    data_test_mmcif_dir = data_test / "mmcifs"
    data_test_clusterings_dir = data_test / "data_caches" / "clusterings"
    data_test_template_dir = data_test / "data_caches" / "template" / "templates"

    if not data_test_mmcif_dir.exists():
        pytest.skip(f"The directory `{data_test_mmcif_dir}` is not populated yet.")

    interface_mapping_path = str(data_test_clusterings_dir / "interface_cluster_mapping.csv")
    chain_mapping_paths = [
        str(data_test_clusterings_dir / "ligand_chain_cluster_mapping.csv"),
        str(data_test_clusterings_dir / "nucleic_acid_chain_cluster_mapping.csv"),
        str(data_test_clusterings_dir / "peptide_chain_cluster_mapping.csv"),
        str(data_test_clusterings_dir / "protein_chain_cluster_mapping.csv"),
    ]

    sampler = WeightedPDBSampler(
        chain_mapping_paths=chain_mapping_paths,
        interface_mapping_path=interface_mapping_path,
        batch_size=64,
    )

    sampler_pdb_ids = set(sampler.mappings.get_column("pdb_id").to_list())
    test_ids = set(
        filepath.stem
        for filepath in data_test_mmcif_dir.glob("**/*.cif")
        if filepath.stem in sampler_pdb_ids
    )

    pdb_input = PDBDataset(
        folder=data_test_mmcif_dir,
        sampler=sampler,
        sample_type="default",
        crop_size=128,
        templates_dir=str(data_test_template_dir),
        sample_only_pdb_ids=test_ids,
        training=False,
    )

    batched_atom_input = pdb_inputs_to_batched_atom_input(pdb_input[0], atoms_per_window=27)
    assert exists(batched_atom_input)
